"""Module providing functions to build an init script

The :func:`mkinit` function will generate a ``/init`` script.

``do_foo()`` functions write a string performing the foo action into a
stream. This stream should be the init script.

``_fun_foo()`` functions write a string defining the foo function into a
stream. This stream should be the init script.

Helper functions available in the init script that can be used by
:class:`cmkinitramfs.data.Data` classes:

 - :func:`_fun_die`: Fatal error handler (does not return).
 - :func:`_fun_log`: Logging functions (always successful).

Special helper function available in the init script:
:func:`_fun_rescue_shell`, :func:`_fun_panic`.

Init environment variables:

 - ``HOME``
 - ``PATH``

Init global variables:

 - ``PRINTK``: Initial kernel log level.
 - ``INIT``: Program to run as init process after the initramfs.
 - ``RD_XXX``: If set, feature *XXX* is enabled.
 - ``RD_BREAK_XXX``: If set, breakpoint *XXX* is enabled.

The init script should never rely on global variables being set or unset,
it should always assign a default value if not set.
"""

from __future__ import annotations

import itertools
import locale
from enum import Enum, auto
from shlex import quote
from typing import Iterable, IO, Mapping, Optional

from .data import Data


#: Global Busybox applet dependencies
BUSYBOX_COMMON_DEPS = {
    '[', 'cat', 'cut', 'echo', 'env', 'exec', 'exit', 'export', 'kill',
    'mount', 'return', 'set', 'shift', 'switch_root', 'sync', 'test',
    'umount', 'uname',
}
#: Keymap loading Busybox applet dependencies
BUSYBOX_KEYMAP_DEPS = {'loadkmap', 'kbd_mode'}
#: Kernel module loading Busybox applet dependencies
BUSYBOX_KMOD_DEPS = {'depmod', 'modprobe'}

#: Get a quoted TAB character
TAB = '"$(printf \'\\t\')"'
#: Get a quoted EOL character
EOL = '"$(printf \'\\n\\b\')"'


class Breakpoint(Enum):
    """Breakpoint in the boot process

    Breakpoints can be enabled by adding rd.break to the kernel command-line
    (e.g. ``./kernel.img foo rd.break=init``).
    Setting ``rd.break=foo,bar`` will enable both ``foo`` and ``bar``.
    Environment variables can also be set to enable them
    (e.g. ``./kernel.img foo RD_BREAK_EARLY=true``).
    """
    #: Early break: break before any action, including command-line parsing.
    #: Can be set with the ``RD_BREAK_EARLY`` environment variable.
    EARLY = auto()
    #: ``init``: Break after initramfs initialization.
    #: Can also be set with the ``RD_BREAK_INIT`` environment variable.
    INIT = auto()
    #: ``module``: Break after loading kernel modules.
    #: Can also be set with the ``RD_BREAK_MODULE`` environment variable.
    #: Alias: ``modules``.
    MODULE = auto()
    #: ``rootfs``: Break after mounting the root filesystem.
    #: Can also be set with the ``RD_BREAK_ROOTFS`` environment variable.
    ROOTFS = auto()
    #: ``mount``: Break after mounting all filesystems.
    #: Can also be set with the ``RD_BREAK_MOUNT`` environment variable.
    #: Alias: ``mounts``.
    MOUNT = auto()


def _fun_rescue_shell(out: IO[str]) -> None:
    """Define the rescue_shell function

    ``rescue_shell`` drop the user to ``/bin/sh``.

    Arguments: none.

    This function *should not* be called from a subshell.

    This function *does not* return.

    :param out: Stream to write into
    """
    out.writelines((
        "rescue_shell()\n",
        "{\n",
        "\tlog 0 'Dropping into a shell'\n",
        "\texec /bin/sh 0<>/dev/console 1<>/dev/console 2<>/dev/console\n",
        "\temerg 'Failed to start rescue shell'\n",
        "\tpanic\n"
        "}\n\n",
    ))


def _fun_panic(out: IO[str]) -> None:
    """Define the panic function

    ``panic`` causes a kernel panic by exiting ``/init``.

    Arguments: none.

    This function *should not* be called from a subshell.

    This function *does not* return.

    :param out: Stream to write into
    """
    out.writelines((
        "panic()\n",
        "{\n",
        "\tlog 0 'Terminating init'\n",
        "\tsync\n",
        "\texit\n",
        "}\n\n",
    ))


def _fun_die(out: IO[str]) -> None:
    """Define the die function

    ``die`` will either start a rescue shell or cause a kernel panic,
    wether ``RD_PANIC`` is set or not.

    Arguments: error message.

    This function *should not* be called from a subshell.

    This function *does not* return.

    :param out: Stream to write into
    """
    out.writelines((
        "die()\n",
        "{\n",
        "\temerg \"$@\"\n",
        "\tkill -TERM -1 || err \'Failed to kill all processes\'\n",
        "\t[ -n \"${RD_PANIC+x}\" ] && panic || rescue_shell\n",
        "}\n\n",
    ))


def _fun_log(out: IO[str]) -> None:
    """Define the logging functions

    ``log``: log a message.

     - Argument 1: syslog level number, from 0 to 7.
     - Additionnal arguments: message to log.

    Logs printed to stderr:

     - Level ≤ 4: always
     - 5 ≤ level ≤ 6: if debug enabled or quiet disabled
     - Level = 7: if debug enabled

    Helper functions:

     - ``emerg``: log a message for a panic condition.
       The message is prepended by 'FATAL:'.
     - ``alert``: log a critical error message requiring immediate action.
       The message is prepended by 'ERROR:'.
     - ``crit``: log a critical error message.
       The message is prepended by 'ERROR:'.
     - ``err``: log an error message.
       The message is prepended by 'ERROR:'.
     - ``warn``: log a warning message.
       The message is prepended by 'WARNING:'.
     - ``notice``: log a significant/unusual informational message.
     - ``info``: log an informational message.
     - ``debug``: log a debug-level message.

    Helper functions will call ``log`` with the coresponding syslog level.

    Logging functions always return successfully.

    :param out: Stream to write into
    """
    out.writelines((
        'log()\n',
        '{\n',
        '\t[ "${1-}" -lt 8 ] && lvl="$1" && shift || lvl=1\n',
        '\t[ $# -ge 1 ] || return 0\n',
        '\techo "<$((24 | lvl))>initramfs:" "$@" 1>/dev/kmsg\n',
        '\tif [ "${lvl}" -eq 5 ] || [ "${lvl}" -eq 6 ] ',
        '&& [ -z "${RD_QUIET+x}" ] || [ -n "${RD_DEBUG+x}" ] ',
        '|| [ "${lvl}" -le 4 ]\n',
        '\tthen echo "$@" 1>&2\n',
        '\tfi\n',
        '\treturn 0\n',
        '}\n',
        '\n',
        'emerg() { log 0 \'FATAL:\' "$@" ; }\n',
        'alert() { log 1 \'ERROR:\' "$@" ; }\n',
        'crit() { log 2 \'ERROR:\' "$@" ; }\n',
        'err() { log 3 \'ERROR:\' "$@" ; }\n',
        'warn() { log 4 \'ERROR:\' "$@" ; }\n',
        'notice() { log 5 "$@" ; }\n',
        'info() { log 6 "$@" ; }\n',
        'debug() { log 7 "$@" ; }\n',
        '\n',
    ))


def do_header(out: IO[str], home: str = '/root', path: str = '/bin:/sbin') \
        -> None:
    """Create the /init header

     - Create the shebang ``/bin/sh``
     - Configure environment variables
     - Define global functions (``panic``, logging, ...)

    :param out: Stream to write into
    :param home: ``HOME`` environment variable
    :param path: ``PATH`` environment variable
    """
    out.writelines((
        "#!/bin/sh\n",
        "\n",
        f"HOME={quote(home)}\n",
        "export HOME\n",
        f"PATH={quote(path)}\n",
        "export PATH\n",
        "\n",
    ))
    _fun_rescue_shell(out)
    _fun_panic(out)
    _fun_die(out)
    _fun_log(out)


def do_init(out: IO[str]) -> None:
    """Initialize the init environment

     - Check the current PID is 1
     - Mount ``/proc``, ``/sys``, ``/dev``
     - Set the kernel log level to 4 (``KERN_ERR`` and higher priority)

    :param out: Stream to write into
    """
    out.writelines((
        "debug 'Initialization'\n",
        "test $$ -eq 1 || die 'init expects to be run as PID 1'\n",
        "mount -t proc none /proc || die 'Failed to mount /proc'\n",
        "mount -t sysfs none /sys || die 'Failed to mount /sys'\n",
        "mount -t devtmpfs none /dev || die 'Failed to mount /dev'\n",
        'PRINTK="$(cut -d', TAB, ' -f1 -s /proc/sys/kernel/printk)"\n',
        "echo 4 1>/proc/sys/kernel/printk || ",
        'err \'Failed to set kernel log level to 4\'\n'
        '[ ! -d "/lib/modules/$(uname -r)" ] || depmod || ',
        'warn \'Failed to generate modules.dep\'\n',
        "\n",
    ))


def do_cmdline(out: IO[str]) -> None:
    """Parse the kernel command line for known parameters

    Note: the command line is parsed up to "--", arguments after this
    are passed through to the final init process.

    Parsed parameters:

     - ``init=<path to init>``: Set the program to run as init process
       after the initramfs.
     - ``debug``: Enable debugging, see ``rd.debug``.
     - ``quiet``: Enable quiet mode, see ``rd.quiet``.
     - ``rd.break={init|rootfs|mount}``: Stops the boot process,
       defaults to ``rootfs``. See :class:`Breakpoint`.
     - ``rd.debug``: Enable debugging mode: output verbose informations.
       If quiet mode is disabled, enable shell trace (with ``set -x``).
     - ``rd.panic``: On fatal error: cause a kernel panic rather than
       dropping into a shell.
     - ``rd.quiet``: Enable quiet mode: reduce verbosity.

    :param out: Stream to write into
    """
    out.writelines((
        "debug 'Parsing command-line'\n",
        "for cmdline in $(cat /proc/cmdline); do\n",
        "\tcase \"${cmdline}\" in\n",
        "\t--) break ;;\n",
        "\tinit=*) INIT=\"${cmdline#*=}\" ;;\n"
        "\tdebug) RD_DEBUG=true ;;\n",
        "\tquiet) RD_QUIET=true ;;\n",
        "\trd.break) RD_BREAK_ROOTFS=true ;;\n",
        "\trd.break=*)\n",
        "\t\told_ifs=\"${IFS}\"\n",
        "\t\tIFS=','\n",
        "\t\tfor bpoint in ${cmdline#*=}; do\n",
        "\t\t\tcase \"${bpoint}\" in\n",
        "\t\t\tinit) RD_BREAK_INIT=true ;;\n",
        "\t\t\tmodule|modules) RD_BREAK_MODULE=true ;;\n",
        "\t\t\trootfs) RD_BREAK_ROOTFS=true ;;\n",
        "\t\t\tmount|mounts) RD_BREAK_MOUNT=true ;;\n",
        "\t\t\t*) err \"Unknown breakpoint ${bpoint}\" ;;\n",
        "\t\t\tesac\n",
        "\t\tdone\n",
        "\t\tIFS=\"${old_ifs}\"\n",
        "\t\t;;\n",
        "\trd.debug) RD_DEBUG=true ;;\n",
        "\trd.panic) RD_PANIC=true ;;\n",
        "\trd.quiet) RD_QUIET=true ;;\n",
        "\t*) unknown_cmd=\"${unknown_cmd-}${unknown_cmd+ }${cmdline}\" ;;\n",
        "\tesac\n",
        "done\n",
        "\n",
        "[ -n \"${RD_DEBUG+x}\" ] && [ -z \"${RD_QUIET+x}\" ] ",
        "&& PS4='+ $0:$LINENO: ' && set -x\n",
        "[ -n \"${unknown_cmd+x}\" ] ",
        "&& debug \"Skipped unknown cmdlines: ${unknown_cmd}\"\n",
        "unset unknown_cmd\n",
        "[ -n \"${RD_DEBUG+x}\" ] || exec 1<>/dev/null || ",
        "err 'Failed to redirect stdout to /dev/null'\n",
        "\n",
    ))


def do_keymap(out: IO[str], keymap_file: str, unicode: bool = True) -> None:
    """Load a keymap

    :param out: Stream to write into
    :param keymap_file: Absolute path of the file to load
    :param unicode: Set the keyboard in unicode mode (rather than ASCII)
    """
    mode = 'unicode' if unicode else 'ASCII'
    out.writelines((
        "info 'Loading keymap'\n",
        f"[ -f {quote(keymap_file)} ] || err ",
        quote(f'Keymap file {keymap_file} not found'), '\n',
        f"kbd_mode {'-u' if unicode else '-a'} || crit ",
        quote(f'Failed to set keyboard mode to {mode}'), '\n',
        f"loadkmap <{quote(keymap_file)} || crit ",
        quote(f'Failed to load keymap {keymap_file}'), '\n',
        "\n",
    ))


def do_module(out: IO[str], module: str, *args: str) -> None:
    """Load a kernel module

    :param out: Stream to write into
    :param module: Name of the module to load
    :param args: Arguments for the module (passed to ``modprobe``)
    """
    quoted_args = (f'{quote(arg)} ' for arg in args)

    out.writelines((
        f"info 'Loading kernel module {module}'\n",
        f"modprobe {quote(module)} ", *quoted_args, '|| crit ',
        quote(f'Failed to load module {module}'), '\n',
        '\n',
    ))


def do_break(out: IO[str], breakpoint_: Breakpoint,
             scripts: Iterable[str] = ()) -> None:
    """Drop into a shell if rd.break is set

    :param out: Stream to write into
    :param breakpoint_: Which breakpoint to check
    :param scripts: User commands to run before the breakpoint
    """
    if breakpoint_ is Breakpoint.EARLY:
        breakname = 'RD_BREAK_EARLY'
    elif breakpoint_ is Breakpoint.INIT:
        breakname = 'RD_BREAK_INIT'
    elif breakpoint_ is Breakpoint.MODULE:
        breakname = 'RD_BREAK_MODULE'
    elif breakpoint_ is Breakpoint.ROOTFS:
        breakname = 'RD_BREAK_ROOTFS'
    elif breakpoint_ is Breakpoint.MOUNT:
        breakname = 'RD_BREAK_MOUNT'
    else:
        raise ValueError(f"Unknown breakpoint: {breakpoint_}")

    if scripts:
        out.write(f"info 'Running user scripts for {breakpoint_}'\n")
        for script in scripts:
            out.writelines((script, "\n"))
        out.write("\n")
    out.writelines((
        "[ -n \"${", breakname, "+x}\" ] && notice ",
        quote(f"Reached {breakpoint_}"), " && rescue_shell\n"
        "\n",
    ))


def do_switch_root(out: IO[str], newroot: Data, init: str = '/sbin/init') \
        -> None:
    """Cleanup and switch root

      - Print debugging information
      - Restore kernel log level (set to boot-time default if not possible)
      - Kill all processes
      - Unmount ``/dev``, ``/sys``, ``/proc``
      - Switch root

    :param out: Stream to write into
    :param newroot: Data to use as new root
    :param init: Init process to execute from the new root
    """
    out.writelines((
        '[ -z "${INIT+x}" ] && INIT=', quote(init), '\n',
        'info "Run ${INIT} as init process"\n',
        'debug \'  with arguments:\'\n',
        'for arg in "${INIT}" "$@"; do debug "    ${arg}"; done\n',
        'debug \'  with environment:\'\n',
        'old_ifs="${IFS}"\n',
        'IFS=', EOL, '\n',
        'for var in $(env); do debug "    ${var}"; done\n',
        'IFS="${old_ifs}"\n',
        '\n',
        '[ -z "${PRINTK+x}" ] && PRINTK=',
        '"$(cut -d', TAB, ' -f4 -s /proc/sys/kernel/printk)"\n',
        'echo "${PRINTK}" 1>/proc/sys/kernel/printk || ',
        'err "Failed to restore kernel log level to ${PRINTK}"\n',
        'exec 0<>/dev/console 1<>/dev/console 2<>/dev/console || ',
        'err \'Failed to restore input/output to console\'\n',
        'kill -TERM -1 || err \'Failed to kill all processes\'\n',
        "umount -l /dev || err 'Failed to unmount /dev'\n",
        "umount -l /proc || err 'Failed to unmount /proc'\n",
        "umount -l /sys || err 'Failed to unmount /sys'\n",
        'exec switch_root ', newroot.path(), ' "${INIT}" "$@"\n',
        "die 'Failed to switch root'\n",
        "\n",
    ))


def mkinit(
        out: IO[str],
        root: Data,
        mounts: Iterable[Data] = (),
        keymap: Optional[str] = None,
        modules: Optional[Mapping[str, Iterable[str]]] = None,
        scripts: Optional[Mapping[Breakpoint, Iterable[str]]] = None,
        ) -> None:  # noqa: E123
    """Create the init script

    :param out: Stream to write into
    :param root: :class:`Data` to use as rootfs
    :param mounts: :class:`Data` needed in addition of rootfs
    :param keymap: Path of the keymap to load, :data:`None` means no keymap
    :param modules: Kernel modules to be loaded in the initramfs:
        ``{module: (arg, ...)}``. ``module`` is the module name string,
        and ``(arg, ...)``` is the iterable with the module parameters.
    :param scripts: User commands to run. ``{breakpoint: commands}``:
        ``breakpoint`` is the :class:`Breakpoint` where the commands will
        be run. ``commands`` is the iterable with the commands.
    """

    datatypes = set()
    for data in itertools.chain((root,), mounts):
        datatypes.add(type(data))
        for dep in data.iter_all_deps():
            datatypes.add(type(dep))
    if modules is None:
        modules = {}
    if scripts is None:
        scripts = {}

    do_header(out)
    do_break(out, Breakpoint.EARLY, scripts.get(Breakpoint.EARLY, ()))
    do_init(out)
    do_cmdline(out)
    if keymap is not None:
        do_keymap(out, keymap,
                  unicode=(locale.getdefaultlocale()[1] == 'UTF-8'))
    for datatype in datatypes:
        datatype.initialize(out)
    do_break(out, Breakpoint.INIT, scripts.get(Breakpoint.INIT, ()))
    for mod_name, mod_args in modules.items():
        do_module(out, mod_name, *mod_args)
    do_break(out, Breakpoint.MODULE, scripts.get(Breakpoint.MODULE, ()))
    root.load(out)
    do_break(out, Breakpoint.ROOTFS, scripts.get(Breakpoint.ROOTFS, ()))
    for mount in mounts:
        mount.load(out)
    do_break(out, Breakpoint.MOUNT, scripts.get(Breakpoint.MOUNT, ()))
    do_switch_root(out, root)



"""Module providing the Data class for init script data sources management

The :class:`Data` class defines an abstract object containing data,
it has multiple subclasses for multiple types of data.
The main methods of those classes are :meth:`Data.load`, :meth:`Data.unload`,
and :meth:`Data.set_final`.

Most functions will write into a stream (:term:`text file`) the content
of the init script.
Use a :class:`io.StringIO` if you need to use strings rather than a stream.
"""

from __future__ import annotations

import itertools
import os.path
from shlex import quote
from typing import Iterable, Iterator, IO, List, Optional, Set, Tuple


class Data:
    """Base class representing any data on the system

    This is an abstract class representing data on the system.
    Its main methods are :meth:`load` and :meth:`unload`.
    :meth:`set_final` declare the object as required for the final
    boot environment (e.g. root fs, usr fs), this will prevent the data
    from being unloaded.

    :param files: Files directly needed in the initramfs.
        Each file is a tuple in the format (``src``, ``dest``),
        where ``src`` is the source file on the current system,
        and ``dest`` is the destination in the initramfs
        (relative to its root directory).
        If ``dest`` is :data:`None`, then ``src`` is used.
    :param execs: Executables directly needed in the initramfs.
        Same format as :attr:`files`.
    :param libs: Libraries directly needed in the initramfs.
        Same format as :attr:`files`.
    :param busybox: Busybox compatible commands needed in the initramfs.
        Any commands that are compatible with Busybox's implementation
        should be added.
        Exception: special shell built-in commands and reserved words
        are guaranteed to be available and *can* be ommitted
        (a list is defined
        in :data:`cmkinitramfs.initramfs.SHELL_SPECIAL_BUILTIN`
        and :data:`cmkinitramfs.initramfs.SHELL_RESERVED_WORDS`).
    :param _need: Loading and runtime dependencies
    :param _lneed: Loading only dependencies
    :param _needed_by: Reverse dependencies
    :param _is_final: The :class:`Data` should not be unloaded
    :param _is_loaded: The :class:`Data` is currently loaded
    """
    files: Set[Tuple[str, Optional[str]]]
    execs: Set[Tuple[str, Optional[str]]]
    libs: Set[Tuple[str, Optional[str]]]
    busybox: Set[str]
    _need: List[Data]
    _lneed: List[Data]
    _needed_by: List[Data]
    _is_final: bool
    _is_loaded: bool

    @classmethod
    def initialize(cls, out: IO[str]) -> None:
        """Initialize the data class

        Initialize the environment for the use of this data class:
        define needed functions and variables.
        A Data class should be initialized only once in the init script,
        before the first :meth:`Data.load` call of the class.

        Default initialization is a no-op and should be redefined
        by subclasses. Subclasses should call their parent :class:`Data`
        class' :meth:`Data.initialize` method.

        :param out: Stream to write into
        """

    def __init__(self) -> None:
        self.files = set()
        self.execs = set()
        self.libs = set()
        self.busybox = set()
        self._need = []
        self._lneed = []
        self._needed_by = []
        self._is_final = False
        self._is_loaded = False

    def __eq__(self, other: object) -> bool:
        return isinstance(other, type(self)) and self.files == other.files \
            and self.execs == other.execs and self.libs == other.libs \
            and self._need == other._need and self._lneed == other._lneed \
            and self._is_final == other._is_final

    def iter_all_deps(self) -> Iterator[Data]:
        """Recursivelly get dependencies

        :return: Iterator over all the dependencies
        """
        for dep in itertools.chain(self._need, self._lneed):
            yield dep
            yield from dep.iter_all_deps()

    def is_final(self) -> bool:
        """Returns a :class:`bool` indicating if the :class:`Data` is final"""
        return self._is_final

    def set_final(self) -> None:
        """This function set the data object as final

        This means the data is required by the final boot environment
        and should never be unloaded (it would be pointless).
        This will also mark its hard dependencies as final.
        """
        self._is_final = True
        for k in self._need:
            k.set_final()

    def add_dep(self, dep: Data) -> None:
        """Add a :class:`Data` object to the hard dependencies"""
        if dep in self._lneed:
            self._lneed.remove(dep)
        if dep not in self._need:
            self._need.append(dep)
        if self not in dep._needed_by:
            dep._needed_by.append(self)

    def add_load_dep(self, dep: Data) -> None:
        """Add a :class:`Data` object to the loading dependencies"""
        if dep not in self._lneed and dep not in self._need:
            self._lneed.append(dep)
        if self not in dep._needed_by:
            dep._needed_by.append(self)

    def _pre_load(self, out: IO[str]) -> None:
        """This function does preparation for loading the Data

        Loads all the needed dependencies.
        It should be called from :meth:`load` before the actual loading
        of the data.
        This method *should not* be called if the :class:`Data` is
        already loaded.

        :param out: Stream to write into
        :raises DataError: Already loaded
        """
        if self._is_loaded:
            raise DataError(f"{self} is already loaded")
        self._is_loaded = True
        # Load dependencies
        for k in self._need + self._lneed:
            if not k._is_loaded:
                k.load(out)

    def _post_load(self, out: IO[str]) -> None:
        """This function does post loading cleanup

        If the object is a loading dependency only, it will load all
        its reverse dependencies in order to be unloaded as soon as possible.
        Unloading quickly can be useful when dealing with sensitive data
        (e.g. a LUKS key).
        It should be called from :meth:`load` after the actual loading
        of the data.

        :param out: Stream to write into
        """
        # If not final, load data needing self, this will allow an
        # unloading as soon as possible
        if not self._is_final:
            for k in self._needed_by:
                if not k._is_loaded:
                    k.load(out)
        # Unload data not needed anymore
        for k in self._lneed:
            k._needed_by.remove(self)
            if not k._needed_by:
                k.unload(out)

    def load(self, out: IO[str]) -> None:
        """This function loads the data

        It should be redefined by subclasses,
        this definition is a no-op only dealing with dependencies.

        Before loading, this function should
        load the dependencies with :meth:`_pre_load`.
        After loading, this function should
        unload unnecessary dependencies with :meth:`_post_load`.
        This method *should not* be called if the data is already loaded.

        :param out: Stream to write into
        """
        self._pre_load(out)
        self._post_load(out)

    def _pre_unload(self, out: IO[str]) -> None:
        """This function does pre unloading sanity checks

        It should be called from :meth:`unload` before the actual unloading
        of the data.

        :param out: Stream to write into
        :raises DataError: Not loaded or dependency issue
        """
        if not self._is_loaded:
            raise DataError(f"{self} is not loaded")
        if self._is_final or self._needed_by:
            raise DataError(f"{self} is still needed or not temporary")

    def _post_unload(self, out: IO[str]) -> None:
        """This function does post unloading cleanup

        It removes itself from the :attr:`_needed_by` reverse dependencies
        of all its dependencies, and check if the dependency can be unloaded.
        This method should be called from :meth:`unload` after the actual
        unloading of the data.
        This *should not* be called if the data is not loaded.

        :param out: Stream to write into
        """
        for k in self._need:
            k._needed_by.remove(self)
            if not k._needed_by:
                k.unload(out)
        self._is_loaded = False

    def unload(self, out: IO[str]) -> None:
        """This function unloads data

        It should be redefined by subclasses,
        this definition is a no-op only dealing with dependencies.

        Before unloading, this function should check for any
        dependency error, with :meth:`_pre_unload`.
        After unloading, this function should unload all unneeded
        dependencies, with :meth:`_post_unload`.

        :param out: Stream to write into
        """
        self._pre_unload(out)
        self._post_unload(out)

    def __str__(self) -> str:
        """Get the name of the data

        This string may be quoted with simple quotes in the script.
        This **has** to be implemented by subclasses.
        """
        raise NotImplementedError()

    def path(self) -> str:
        """Get the path of this data

        This function provides a string allowing access to data from within
        the init environment, this string can be a path or a command
        in a subshell (e.g. ``"$(findfs UUID=foobar)"``).
        This string should be ready to be used in the script without
        being quoted nor escaped.
        This **has** to be implemented by subclasses.
        """
        raise NotImplementedError()


class DataError(Exception):
    """Error in the :class:`Data` object"""


class PathData(Data):
    """Absolute path

    :param datapath: Path of the data
    """
    datapath: str

    def __init__(self, datapath: str):
        super().__init__()
        self.datapath = datapath

    def __str__(self) -> str:
        return self.datapath

    def __eq__(self, other: object) -> bool:
        return isinstance(other, type(self)) and super().__eq__(other) \
            and self.datapath == other.datapath

    def path(self) -> str:
        return quote(self.datapath)


class UuidData(Data):
    """UUID of a data

    The UUID can be a filesystem UUID, or other UUID known by other
    :class:`Data` classes (e.g. a MD UUID).

    :param uuid: UUID of the data
    :param partition: If :data:`True`, the UUID is treated as a partition UUID
    """
    uuid: str
    partition: bool

    def __init__(self, uuid: str, partition: bool = False):
        super().__init__()
        if partition:
            # PARTUUID is only available in util-linux findfs
            self.execs |= {('findfs', None)}
        else:
            self.busybox |= {'findfs'}
        self.uuid = uuid
        self.partition = partition

    def __str__(self) -> str:
        return ('PARTUUID=' if self.partition else 'UUID=') + self.uuid

    def __eq__(self, other: object) -> bool:
        return isinstance(other, type(self)) and super().__eq__(other) \
            and self.uuid == other.uuid and self.partition == other.partition

    def path(self) -> str:
        prefix = 'PARTUUID=' if self.partition else 'UUID='
        return '"$(findfs ' + quote(prefix + self.uuid) + ')"'


class LabelData(Data):
    """Label of a data

    The label can be a filesystem or partition label, or a label known
    by other :class:`Data` classes.

    :param label: Label of the data
    :param partition: If :data:`True`, the label is treated as a partition
        label
    """
    label: str
    partition: bool

    def __init__(self, label: str, partition: bool = False):
        super().__init__()
        if partition:
            # PARTLABEL is only available in util-linux findfs
            self.execs |= {('findfs', None)}
        else:
            self.busybox |= {'findfs'}
        self.label = label
        self.partition = partition

    def __str__(self) -> str:
        return ('PARTLABEL=' if self.partition else 'LABEL=') + self.label

    def __eq__(self, other: object) -> bool:
        return isinstance(other, type(self)) and super().__eq__(other) \
            and self.label == other.label and self.partition == other.partition

    def path(self) -> str:
        prefix = 'PARTLABEL=' if self.partition else 'LABEL='
        return '"$(findfs ' + quote(prefix + self.label) + ')"'


class LuksData(Data):
    """LUKS encrypted block device

    :param source: :class:`Data` to unlock (crypto_LUKS volume),
        it will be set as a hard dependency
    :param name: Name for the LUKS volume
    :param key: :class:`Data` to use as key file,
        it will be set as a load dependency
    :param header: :class:`Data` containing the LUKS header,
        it will be set as a load dependency
    :param discard: Enable discards
    """
    source: Data
    name: str
    key: Optional[Data]
    header: Optional[Data]
    discard: bool

    def __init__(self, source: Data, name: str,
                 key: Optional[Data] = None, header: Optional[Data] = None,
                 discard: bool = False):
        super().__init__()
        self.execs.add(('cryptsetup', None))
        self.libs.add(('libgcc_s.so.1', None))
        self.source = source
        self.name = name
        self.key = key
        self.header = header
        self.discard = discard
        self.add_dep(self.source)
        if self.key:
            self.add_load_dep(self.key)
        if self.header:
            self.add_load_dep(self.header)

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        return isinstance(other, type(self)) and super().__eq__(other) \
            and self.source == other.source and self.name == other.name \
            and self.key == other.key and self.header == other.header \
            and self.discard == other.discard

    def load(self, out: IO[str]) -> None:
        header = f'--header {self.header.path()} ' if self.header else ''
        key_file = f'--key-file {self.key.path()} ' if self.key else ''
        discard = '--allow-discards ' if self.discard else ''
        self._pre_load(out)
        out.writelines((
            f"info 'Unlocking LUKS device {self}'\n",
            "cryptsetup ", header, key_file, discard,
            f"open {self.source.path()} {quote(self.name)} || die ",
            quote(f'Failed to unlock LUKS device {self}'), '\n',
            "\n",
        ))
        self._post_load(out)

    def unload(self, out: IO[str]) -> None:
        self._pre_unload(out)
        out.writelines((
            f"info 'Closing LUKS device {self}'\n",
            f"cryptsetup close {quote(self.name)} || die ",
            quote(f'Failed to close LUKS device {self}'), '\n',
            "\n",
        ))
        self._post_unload(out)

    def path(self) -> str:
        return quote('/dev/mapper/' + self.name)


class LvmData(Data):
    """LVM logical volume

    :param vg_name: Name of the volume group
    :param lv_name: Name of the logical volume
    """
    vg_name: str
    lv_name: str

    @staticmethod
    def __lvm_conf(out: IO[str]) -> None:
        """Create LVM config in ``/etc/lvm/lvmlocal.conf``

        This override some configurations specific to the
        initramfs environment.

        Note: if ``/etc/lvm/lvmlocal.conf`` exists, we append to it,
        which may cause duplicate configuration warnings from LVM.
        """
        out.writelines((
            "debug 'Writing LVM configuration'\n",
            "mkdir -p /etc/lvm && touch /etc/lvm/lvmlocal.conf || warn ",
            "'Failed to create LVM configuration file'\n",
            "{\n",
            "\techo 'activation/monitoring = 0'\n",
            "\techo 'activation/udev_rules = 0'\n",
            "\techo 'activation/udev_sync = 0'\n",
            "\techo 'devices/external_device_info_source = \"none\"'\n",
            "\techo 'devices/md_component_detection = 0'\n",
            "\techo 'devices/multipath_component_detection = 0'\n",
            "\techo 'devices/obtain_device_list_from_udev = 0'\n",
            "\techo 'global/locking_type = 4'\n",
            "\techo 'global/use_lvmetad = 0'\n",
            "\techo 'global/use_lvmlockd = 0'\n",
            "\techo 'global/use_lvmpolld = 0'\n",
            "} >>/etc/lvm/lvmlocal.conf || warn ",
            "'Failed to write LVM configuration file'\n"
            "\n",
        ))

    @classmethod
    def initialize(cls, out: IO[str]) -> None:
        super().initialize(out)
        LvmData.__lvm_conf(out)

    def __init__(self, vg_name: str, lv_name: str):
        super().__init__()
        self.execs.add(('lvm', None))
        self.vg_name = vg_name
        self.lv_name = lv_name

    def __str__(self) -> str:
        return self.vg_name + "/" + self.lv_name

    def __eq__(self, other: object) -> bool:
        return isinstance(other, type(self)) and super().__eq__(other) \
            and self.vg_name == other.vg_name and self.lv_name == other.lv_name

    def load(self, out: IO[str]) -> None:
        self._pre_load(out)
        out.writelines((
            f"info 'Enabling LVM logical volume {self}'\n",
            "lvm lvchange --sysinit -a ly ",
            f"{quote(f'{self.vg_name}/{self.lv_name}')} || die ",
            quote(f'Failed to enable LVM logical volume {self}'), '\n',
            "lvm vgmknodes || err ",
            quote(f'Failed to create LVM nodes for {self}'), '\n',
            "\n",
        ))
        self._post_load(out)

    def unload(self, out: IO[str]) -> None:
        self._pre_unload(out)
        out.writelines((
            f"info 'Disabling LVM logical volume {self}'\n",
            "lvm lvchange --sysinit -a ln ",
            f"{quote(f'{self.vg_name}/{self.lv_name}')} || die ",
            quote(f'Failed to disable LVM logical volume {self}'), '\n',
            "lvm vgmknodes || err ",
            quote(f'Failed to remove LVM nodes for {self}'), '\n',
            "\n",
        ))
        self._post_unload(out)

    def path(self) -> str:
        # If LV or VG name has an hyphen '-', LVM doubles it in the path
        return quote('/dev/mapper/' + self.vg_name.replace('-', '--')
                     + '-' + self.lv_name.replace('-', '--'))


class MountData(Data):
    """Mount point

    :param source: :class:`Data` to use as source
        (e.g. /dev/sda1, my-luks-data),
        it will be set as a hard dependency
    :param mountpoint: Absolute path of the mountpoint
    :param filesystem: Filesystem (used for ``mount -t filesystem``)
    :param options: Mount options
    """
    source: Data
    mountpoint: str
    filesystem: str
    options: str

    @staticmethod
    def __fun_fsck(out: IO[str]) -> None:
        """Define the mount_fsck function

        This function takes any number of arguments, which will be passed
        to ``fsck``. This function checks the return code of the ``fsck``
        command and acts accordingly.

        This functions calls ``fsck`` with all the arguments it was called
        with. It checks the return code of ``fsck`` and :

        - No error: returns 0.
        - Non fatal error: prints an error and returns 0.
        - Non fatal error requiring reboot: prints an error and reboot.
        - Fatal error: returns 1.

        :param out: Stream to write into
        """
        fsck_err = {
            1: ('notice', "Filesystem errors corrected"),
            2: ('notice', "System should be rebooted"),
            4: ('alert', "Filesystem errors left uncorrected"),
            8: ('crit', "Operational error"),
            16: ('crit', "Usage or syntax error"),
            32: ('err', "Checking canceled by user request"),
            128: ('crit', "Shared-library error"),
        }
        code_err = 4 | 8 | 16 | 32 | 64 | 128
        code_reboot = 2

        out.writelines((
            'mount_fsck()\n',
            '{\n',
            '\tFSTAB_FILE=/dev/null fsck "$@"\n',
            '\tfsck_ret=$?\n'
            '\t[ "${fsck_ret}" -eq 0 ] && return 0\n',
        ))
        for err_code in fsck_err:
            err_call, err_str = fsck_err[err_code]
            out.writelines((
                f'\t[ "$((fsck_ret & {err_code}))" -eq {err_code} ] && ',
                err_call, ' ', quote(f"fsck: {err_str}"), '\n',
            ))
        out.writelines((
            '\t[ "$((fsck_ret & ', str(code_err), '))" -ne 0 ] && return 1\n',
            '\tif [ "$((fsck_ret & ', str(code_reboot), '))" -eq 2 ]; then ',
            'notice \'Rebooting...\'; reboot -f; fi\n',
            '\treturn 0\n',
            '}\n',
            '\n',
        ))

    @classmethod
    def initialize(cls, out: IO[str]) -> None:
        super().initialize(out)
        MountData.__fun_fsck(out)

    def __init__(self, source: Data, mountpoint: str, filesystem: str,
                 options: str = "ro"):
        super().__init__()
        self.busybox |= {'fsck', '[', 'reboot', 'mkdir', 'mount', 'umount'}
        self.source = source if source else PathData("none")
        self.mountpoint = mountpoint
        self.filesystem = filesystem
        self.options = options
        if self.filesystem in ('btrfs',):
            self.execs.add(('btrfs', None))
            self.execs.add(('fsck.btrfs', None))
        elif self.filesystem in ('ext4',):
            self.execs.add(('fsck.ext4', None))
            self.execs.add(('e2fsck', None))
        elif self.filesystem in ('xfs',):
            self.execs.add(('fsck.xfs', None))
            self.execs.add(('xfs_repair', None))
        elif self.filesystem in ('fat', 'vfat'):
            self.execs.add(('fsck.fat', None))
            self.execs.add(('fsck.vfat', None))
        elif self.filesystem in ('exfat',):
            self.execs.add(('fsck.exfat', None))
        elif self.filesystem in ('f2fs',):
            self.execs.add(('fsck.f2fs', None))
        self.add_dep(self.source)

    def __str__(self) -> str:
        return self.mountpoint

    def __eq__(self, other: object) -> bool:
        return isinstance(other, type(self)) and super().__eq__(other) \
            and self.source == other.source \
            and self.mountpoint == other.mountpoint \
            and self.filesystem == other.filesystem \
            and self.options == other.options

    def load(self, out: IO[str]) -> None:
        fsck = (
            f'mount_fsck -t {quote(self.filesystem)} ',
            f'{self.source.path()} || die ',
            quote(f'Failed to check filesystem {self}'), '\n',
        ) if self.source.path() != 'none' else ()
        mkdir = (
            f"[ -d {quote(self.mountpoint)} ] || ",
            f"mkdir {quote(self.mountpoint)} || err ",
            quote(f'Failed to create directory {self}'), '\n',
        ) if os.path.dirname(self.mountpoint) == '/mnt' else ()

        self._pre_load(out)
        out.writelines((
            f"info 'Mounting filesystem {self}'\n",
            *fsck,
            *mkdir,
            f"mount -t {quote(self.filesystem)} -o {quote(self.options)} ",
            f"{self.source.path()} {quote(self.mountpoint)} || die ",
            quote(f'Failed to mount filesystem {self}'), '\n',
            "\n",
        ))
        self._post_load(out)

    def unload(self, out: IO[str]) -> None:
        self._pre_unload(out)
        out.writelines((
            f"info 'Unmounting filesystem {self}'\n",
            f"umount {quote(self.mountpoint)} || die ",
            quote(f'Failed to unmount filesystem {self}'), '\n',
            "\n",
        ))
        self._post_unload(out)

    def path(self) -> str:
        return quote(self.mountpoint)


class MdData(Data):
    """MD RAID

    :param sources: :class:`Data` to use as sources (e.g. /dev/sda1 and
        /dev/sdb1; or UUID=foo), they will be set as hard dependencies
    :param name: Name for the MD device
    :raises ValueError: No :class:`Data` source
    """
    sources: Tuple[Data, ...]
    name: str

    def __init__(self, sources: Iterable[Data], name: str):
        super().__init__()
        self.execs.add(('mdadm', None))
        self.sources = tuple(sources)
        self.name = name
        if not self.sources:
            raise ValueError(f"{self} has no source defined")
        for source in self.sources:
            self.add_dep(source)

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        return isinstance(other, type(self)) and super().__eq__(other) \
            and self.sources == other.sources and self.name == other.name

    def load(self, out: IO[str]) -> None:
        # Get the string containing all sources to use
        sources: Set[str] = set()
        for source in self.sources:
            if isinstance(source, UuidData):
                sources.add(f"--uuid {quote(source.uuid)} ")
            else:
                sources.add(f"{source.path()} ")

        self._pre_load(out)
        out.writelines((
            f"info 'Assembling MD RAID {self}'\n",
            "MDADM_NO_UDEV=1 ",
            "mdadm --assemble ", *sources, f"{quote(self.name)} || die ",
            quote(f'Failed to assemble MD RAID {self}'), '\n',
            "\n",
        ))
        self._post_load(out)

    def unload(self, out: IO[str]) -> None:
        self._pre_unload(out)
        out.writelines((
            f"info 'Stopping MD RAID {self}'\n",
            "MDADM_NO_UDEV=1 ",
            f"mdadm --stop {quote(self.name)} || die ",
            quote(f'Failed to stop MD RAID {self}'), '\n',
            "\n",
        ))
        self._post_unload(out)

    def path(self) -> str:
        return quote('/dev/md/' + self.name)


class CloneData(Data):
    """Clone a :class:`Data` to another

    :param source: :class:`Data` to use as source,
        it will be set as a load dependency
    :param dest: :class:`Data` to use as destination,
        it will be set as a hard dependency
    """
    source: Data
    dest: Data

    def __init__(self, source: Data, dest: Data):
        super().__init__()
        self.busybox |= {'cp'}
        self.source = source
        self.dest = dest
        self.add_load_dep(self.source)
        self.add_dep(self.dest)

    def __str__(self) -> str:
        return f"{self.source} to {self.dest}"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, type(self)) and super().__eq__(other) \
            and self.source == other.source and self.dest == other.dest

    def load(self, out: IO[str]) -> None:
        self._pre_load(out)
        out.writelines((
            f"info 'Cloning {self}'\n",
            f"cp -aT {self.source.path()} {self.dest.path()} || die ",
            quote(f'Failed to clone {self}'), '\n',
            "\n",
        ))
        self._post_load(out)

    def path(self) -> str:
        return self.dest.path()



"""Module providing functions and classes to build an initramfs

This library provides a class :class:`Initramfs` which handles the content
of an initramfs. This class supports creating an initramfs tree inside
a directlry. It also can generate a CPIO file list compatible with the Linux
kernel's ``gen_init_cpio`` utility. (See
https://www.kernel.org/doc/html/latest/filesystems/ramfs-rootfs-initramfs.html
for mor details.)

The main function is :func:`mkinitramfs` to build the complete initramfs.
"""

from __future__ import annotations

import logging
import os
import os.path
import platform
import subprocess
from typing import IO, Iterable, Iterator, List, Optional, Set, Tuple

from .bin import (find_elf_deps_set, find_kmod, find_kmod_deps,
                  find_exec, find_lib)
from .item import Directory, File, Item, MergeError, Node, Symlink
from .utils import hash_file, normpath, removeprefix

logger = logging.getLogger(__name__)
#: Set of shell special built-in commands.
#: They are guaranteed to be available in the initramfs' ``/bin/sh``.
SHELL_SPECIAL_BUILTIN = frozenset((
    'break', ':', 'continue', '.', 'eval', 'exec', 'exit', 'export',
    'readonly', 'return', 'set', 'shift', 'times', 'trap', 'unset',
))
#: Set of shell reserved words.
#: They are guaranteed to be available in the initramfs' ``/bin/sh``.
SHELL_RESERVED_WORDS = frozenset((
    '!', '{', '}', 'case', 'do', 'done', 'elif', 'else', 'esac', 'fi', 'for',
    'if', 'in', 'then', 'until', 'while',
))


def busybox_get_applets(busybox_exec: str) -> Iterator[str]:
    """Get BusyBox applets

    :param busybox_exec: BusyBox executable (e.g. ``busybox``)
    :return: Iterator of absolute paths of BusyBox applets
    :raises subprocess.CalledProcessError: Error during ``busybox_exec``
    """
    cmd = (busybox_exec, '--list-full')
    logger.debug("Subprocess: %s", cmd)
    with subprocess.Popen(cmd, stdout=subprocess.PIPE) as proc:
        assert proc.stdout is not None
        for line in proc.stdout:
            yield '/' + line.decode().strip()
        if proc.wait() != 0:
            raise subprocess.CalledProcessError(proc.returncode, proc.args)


def mkcpio_from_dir(src: str, dest: IO[bytes]) -> None:
    """Create CPIO archive from a given directory

    :param src: Directory from which the archive is created
    :param dest: Destination stream of the CPIO data
    :raises subprocess.CalledProcessError: Error during ``find`` or ``cpio``
    """
    logger.debug("Creating CPIO archive")

    oldpwd = os.getcwd()
    os.chdir(src)
    cmd: Tuple[str, ...] = ("find", ".", "-print0")
    logger.debug("Subprocess: %s", cmd)
    with subprocess.Popen(cmd, stdout=subprocess.PIPE) as find:
        cmd = ('cpio', '--quiet', '--null', '--create', '--format=newc')
        logger.debug("Subprocess: %s", cmd)
        with subprocess.Popen(cmd, stdin=find.stdout, stdout=dest) as cpio:
            if cpio.wait() != 0:
                raise subprocess.CalledProcessError(cpio.returncode, cpio.args)
        if find.wait() != 0:
            raise subprocess.CalledProcessError(find.returncode, find.args)
    os.chdir(oldpwd)


def mkcpio_from_list(src: str, dest: IO[bytes]) -> None:
    """Create CPIO archive from a given CPIO list

    :param src: Path of the CPIO list
    :param dest: Destination stream of the CPIO data
    :raises subprocess.CalledProcessError: Error during ``gen_init_cpio``
    """
    cmd = ('gen_init_cpio', src)
    logger.debug("Subprocess: %s", cmd)
    subprocess.check_call(cmd, stdout=dest)


def keymap_build(src: str, dest: IO[bytes], unicode: bool = True) -> None:
    """Generate a binary keymap from a keymap name

    This keymap can then be loaded with ``loadkmap`` from the initramfs
    environment.

    :param src: Name of the keymap to convert, can be a keyboard layout name
        or a file path
    :param dest: Destination stream to write into
    :param unicode: Generate a unicode keymap (rather than ASCII)
    :raises subprocess.CalledProcessError: Error during ``loadkeys``
    """
    cmd = ('loadkeys', '--unicode' if unicode else '--ascii', '--bkeymap', src)
    logger.debug("Subprocess: %s", cmd)
    subprocess.check_call(cmd, stdout=dest)


class Initramfs:
    """An initramfs archive

    :param user: Default user to use when creating items
    :param group: Default group to use when creating items
    :param binroot: Root directory where binary files are found
        (executables and libraries)
    :param kernels: Kernel versions of the initramfs,
        defaults to the running kernel version
    :param items: Items in the initramfs
    """
    user: int
    group: int
    binroot: str
    kernels: Set[str]
    items: List[Item]

    def __init__(self, user: int = 0, group: int = 0, binroot: str = '/',
                 kernels: Optional[Iterable[str]] = None) -> None:
        self.user = user
        self.group = group
        self.binroot = binroot
        self.kernels = set(kernels) if kernels is not None \
            else {platform.release()}
        logger.debug("Target kernels: %s", self.kernels)
        self.items = []
        self.__mklayout()

    def __mklayout(self) -> None:
        """Create the base layout of the initramfs"""
        logger.debug("Creating initramfs layout")

        self.items.append(Directory(0o755, self.user, self.group, '/'))

        # Base layout
        self.add_item(Directory(0o755, self.user, self.group, '/bin'))
        self.add_item(Directory(0o755, self.user, self.group, '/dev'))
        self.add_item(Directory(0o755, self.user, self.group, '/etc'))
        self.add_item(Directory(0o755, self.user, self.group, '/mnt'))
        self.add_item(Directory(0o755, self.user, self.group, '/proc'))
        self.add_item(Directory(0o755, self.user, self.group, '/root'))
        self.add_item(Directory(0o755, self.user, self.group, '/run'))
        self.add_item(Directory(0o755, self.user, self.group, '/sbin'))
        self.add_item(Directory(0o755, self.user, self.group, '/sys'))

        # Only create /lib* if they exists on the current system
        for libdir in ["/lib", "/lib32", "/lib64"]:
            if os.path.islink(libdir):
                self.add_item(Symlink(0o777, self.user, self.group,
                                      libdir, os.readlink(libdir)))
            elif os.path.isdir(libdir):
                self.add_item(Directory(0o755, self.user, self.group, libdir))

        # Create necessary character devices
        self.add_item(Node(0o600, self.user, self.group, '/dev/console',
                           Node.NodeType.CHARACTER, 5, 1))
        self.add_item(Node(0o666, self.user, self.group, '/dev/tty',
                           Node.NodeType.CHARACTER, 5, 0))
        self.add_item(Node(0o666, self.user, self.group, '/dev/null',
                           Node.NodeType.CHARACTER, 1, 3))
        self.add_item(Node(0o644, self.user, self.group, '/dev/kmsg',
                           Node.NodeType.CHARACTER, 1, 11))

        # Add kernel modules information
        for kernel in self.kernels:
            kmod_dir = f'/lib/modules/{kernel}'
            self.mkdir(kmod_dir, mode=0o755, parents=True)
            self.add_file(kmod_dir + '/modules.order', mode=0o640)
            self.add_file(kmod_dir + '/modules.builtin', mode=0o640)
            self.add_file(kmod_dir + '/modules.builtin.modinfo', mode=0o640)

    def __iter__(self) -> Iterator[Item]:
        """Iterate over the :class:`Item` instances in the initramfs"""
        return iter(self.items)

    def __contains__(self, path: str) -> bool:
        """Check if a path exists on the initramfs

        :param path: Path to check
        :return: :data:`True` if ``path`` exists on the initramfs,
            :data:`False` otherwise
        """
        for item in self:
            if path in item:
                return True
        return False

    def __eq__(self, other: object) -> bool:
        return isinstance(other, type(self)) and self.user == other.user \
            and self.group == other.group and self.binroot == other.binroot \
            and self.kernels == other.kernels and self.items == other.items

    def add_item(self, new_item: Item) -> None:
        """Add an item to the initramfs

        If an identical item is already present, merges them together.

        :param new_item: :class:`Item` instance to add
        :raises MergeError: Item cannot be merged into the initramfs
            (missing parent directory or file conflict)
        """

        #: Used to check all parents directories exist and are created
        #: before the creation/merge of new_item
        parents = {os.path.dirname(k): False for k in new_item if k != '/'}
        mergeable = None
        for cur_item in self:
            # Check if new_item can be merged or is conflicting with cur_item
            if cur_item.is_mergeable(new_item):
                assert mergeable is None
                mergeable = cur_item
            for dest in new_item:
                if cur_item is not mergeable and dest in cur_item:
                    raise MergeError(
                        f"File collision between {new_item} and {cur_item}"
                    )
                if mergeable is None and os.path.dirname(dest) in cur_item:
                    parents[os.path.dirname(dest)] = True

        if not all(parents.values()):
            missings = tuple(k for k in parents if not parents[k])
            logger.error("Cannot add %s: missing directories %s",
                         new_item, missings)
            raise MergeError(f"Missing directory: {missings}")
        if mergeable is not None:
            mergeable.merge(new_item)
        else:
            # Add new_item
            self.items.append(new_item)
            logger.debug("New item: %s", new_item)

    @staticmethod
    def __normalize(path: str) -> str:
        """Normalize a path for the initramfs filesystem

        Strip /usr[/local] directory, warns if spaces are present in
        the path.

        :param path: Destination path to normalize
        :return: Normalized path
        :raises ValueError: Invalid path
        """

        path = normpath(path)
        # Initramfs path must be absolute
        if not os.path.isabs(path):
            raise ValueError(f"{path} is not an absolute path")
        # Strip /usr directory, not needed in initramfs
        if path.startswith('/usr/local/'):
            logger.debug("Stripping /usr/local/ from %s", path)
            path = removeprefix(path, '/usr/local')
        elif path.startswith('/usr/'):
            logger.debug("Stripping /usr/ from %s", path)
            path = removeprefix(path, '/usr')
        # Check whitespaces
        if len(path.split()) != 1:
            logger.warning("Whitespaces are not supported by gen_init_cpio: "
                           "%s", path)
        return path

    def mkdir(self, path: str, mode: int = 0o755, parents: bool = False) \
            -> None:
        """Create a directory on the initramfs

        :param path: Absolute path of the directory,
            relative to the initramfs root
        :param mode: File permissions to use
        :param parents: If :data:`True`, missing parent directories will
            also be created
        :raises MergeError: Destination file exists and is different,
            or missing parent directory (raised from :meth:`add_item`)
        """
        logger.debug("Creating directory %s", path)
        if parents and os.path.dirname(path) not in self:
            self.mkdir(os.path.dirname(path), mode=mode, parents=True)
        self.add_item(Directory(mode, self.user, self.group, path))

    def add_file(self, src: str, dest: Optional[str] = None,
                 mode: Optional[int] = None) -> None:
        """Add a file to the initramfs

        If the file is a symlink, it is dereferenced.
        If it is a dynamically linked ELF file, its dependencies
        are also added.

        :param src: Absolute or relative path of the source file
        :param dest: Absolute path of the destination, relative to the
            initramfs root, defaults to ``src``
        :param mode: File permissions to use, defaults to same as ``src``
        :raises FileNotFoundError: Source file or ELF dependency not found
        :raises MergeError: Destination file exists and is different,
            or missing parent directory (raised from :meth:`add_item`)
        """

        # Sanity checks
        if not os.path.exists(src):
            raise FileNotFoundError(src)

        # Configure paths
        src = os.path.abspath(src)
        if not dest:
            dest = src
        dest = Initramfs.__normalize(dest)

        logger.debug("Adding %s as %s", src, dest)

        # Copy dependencies
        for dep_src, dep_dest in find_elf_deps_set(src, self.binroot):
            self.add_file(dep_src, dep_dest)

        # Add file
        if mode is None:
            mode = os.stat(src, follow_symlinks=True).st_mode & 0o7777
        self.add_item(File(mode, self.user, self.group,
                           {dest}, src, hash_file(src)))

    def add_library(self, src: str, dest: Optional[str] = None,
                    mode: Optional[int] = None) -> None:
        """Add a library to the initramfs

        :param src: Path or base name of the library to add,
            if it is not a path, it is searched on the system with
            :func:`find_lib`
        :param dest: Absolute path of the destination, relative to the
            initramfs root, defaults to the path of the source library
        :param mode: File permissions to use, defaults to same as ``src``
        :raises FileNotFoundError: Library or ELF dependency not found
        :raises MergeError: Destination file exists and is different,
            or missing parent directory (raised from :meth:`add_item`)
        """
        lib_src, lib_dest = find_lib(src, root=self.binroot)
        self.add_file(lib_src, dest if dest is not None else lib_dest,
                      mode=mode)

    def add_executable(self, src: str, dest: Optional[str] = None,
                       mode: Optional[int] = None) -> None:
        """Add an executable to the initramfs

        :param src: Path or base name of the executable to add,
            if it is not a path, it is searched on the system with
            :func:`find_exec`
        :param dest: Absolute path of the destination, relative to the
            initramfs root, defaults to the path of the source executable
        :param mode: File permissions to use, defaults to same as ``src``
        :raises FileNotFoundError: Executable or ELF dependency not found
        :raises MergeError: Destination file exists and is different,
            or missing parent directory (raised from :meth:`add_item`)
        """
        exec_src, exec_dest = find_exec(src, root=self.binroot)
        self.add_file(exec_src, dest if dest is not None else exec_dest,
                      mode=mode)

    def add_kmod(self, module: str, mode: Optional[int] = None) -> None:
        """Add a kernel module to the initramfs

        :param module: Path or name of the kernel module to add
        :param mode: File permissions to use, defaults to same as ``module``
        """

        def _add_kmod(module: str, kernel: str) -> None:
            kmod = find_kmod(module, kernel)
            for dep in find_kmod_deps(kmod):
                _add_kmod(dep, kernel)
            self.mkdir(os.path.dirname(kmod), parents=True)
            self.add_file(kmod, mode=mode)

        for kernel in self.kernels:
            _add_kmod(module, kernel)

    def add_busybox(self, needed: Iterable[str] = (),
                    sys_busybox: Optional[str] = None) -> None:
        """Add busybox and its applets to the initramfs

        Applets will be ignored if a file with the same path
        already exists. To avoid collision, this method should be called
        after all needed files have been added.

        If any command listed in ``needed`` is not available in busybox
        (either as an applet, a shell special built-in, or a reserved word),
        the default system one will be included (from ``PATH``).

        :param needed: Needed Busybox compatible shell commands
        :param sys_busybox: Busybox executable to use to get the list of
            applets, defaults to the one in ``PATH``
        :raises FileNotFoundError: Busybox executable or ELF dependency
            not found
        :raises MergeError: Destination file exists and is different,
            or missing parent directory (raised from :meth:`add_item`,
            not raised for applets)
        """
        if sys_busybox is None:
            sys_busybox = find_exec('busybox')[0]
        applets = set() | SHELL_SPECIAL_BUILTIN | SHELL_RESERVED_WORDS

        busybox_src, busybox_dest = find_exec('busybox', root=self.binroot)
        self.add_file(busybox_src, busybox_dest)
        for applet in busybox_get_applets(sys_busybox):
            applets.add(os.path.basename(applet))
            try:
                self.add_file(busybox_src, applet)
            except MergeError:
                logger.debug("Not adding applet %s: file exists", applet)
        for dep in needed:
            if dep not in applets:
                logger.debug("Adding missing applet: %s", dep)
                self.add_executable(dep)

    def build_to_cpio_list(self, dest: IO[str]) -> None:
        """Write a CPIO list into a file

        This list is compatible with Linux's ``gen_init_cpio``.
        See :meth:`Item.build_to_cpio_list`.

        :param dest: Stream in which the list is written
        """
        for item in self:
            logger.debug("Outputting %s", item)
            dest.write(item.build_to_cpio_list())
            dest.write('\n')

    def build_to_directory(self, dest: str, do_nodes: bool = True) -> None:
        """Copy or create all items to a real filesystem

        See :meth:`Item.build_to_directory`.

        :param dest: Path to use as root directory of the initramfs
        :param do_nodes: Also creates :class:`Node` items, (used for debugging:
            ``CAP_MKNOD`` is needed to create some special devices)
        """
        for item in self:
            if not do_nodes and isinstance(item, Node):
                logger.warning("Not building Node %s", item)
                continue
            logger.debug("Building %s", item)
            item.build_to_directory(dest)



"""Module providing miscellaneous utilities used by cmkinitramfs"""

from __future__ import annotations

import functools
import hashlib
import os.path


# Function needed for python < 3.9
def removeprefix(string: str, prefix: str) -> str:
    """Remove a prefix from a string

    Add support for :meth:`str.removeprefix` for Python < 3.9.

    :param string: String to remove prefix from
    :param prefix: Prefix to remove
    """
    # return string.removeprefix(prefix)
    if string.startswith(prefix):
        return string[len(prefix):]
    return string


def normpath(path: str) -> str:
    """Normalize path (actually eliminates double slashes)

    :param path: Path to normalize
    """
    return os.path.normpath(path).replace('//', '/')


@functools.lru_cache()
def hash_file(filepath: str, chunk_size: int = 65536) -> bytes:
    """Calculate the SHA512 of a file

    :param filepath: Path of the file to hash
    :param chunk_size: Number of bytes per chunk of file to hash
    :return: File hash in a :class:`bytes` object
    """
    sha512 = hashlib.sha512()
    with open(filepath, 'rb') as src:
        for chunk in iter(lambda: src.read(chunk_size), b''):
            sha512.update(chunk)
    return sha512.digest()



"""Module providing functions to manage binaries and executables

This library also provides utilities like :func:`find_exec`.
Multiple helper functions are also available (e.g. :func:`find_elf_deps_iter`
and :func:`find_elf_deps_set` to list libraries needed by an ELF executable).
"""

from __future__ import annotations

import functools
import glob
import itertools
import logging
import os
import os.path
import platform
import subprocess
from typing import FrozenSet, Iterator, List, Optional, Tuple, Union

from elftools.common.exceptions import ELFError
from elftools.elf.elffile import ELFFile
from elftools.elf.enums import ENUM_DT_FLAGS_1

from .utils import normpath, removeprefix


logger = logging.getLogger(__name__)
#: Kernel modules will be searched in ``{KMOD_DIR}/{KERNEL}/**/*.ko``
KMOD_DIR = '/lib/modules'


class ELFIncompatibleError(ELFError):
    """The ELF files are incompatible"""


def parse_ld_path(ld_path: Optional[str] = None, origin: str = '',
                  root: str = '/') -> Iterator[str]:
    """Parse a colon-delimited list of paths and apply ldso rules

    Note the special handling as dictated by the ldso:
     - Empty paths are equivalent to $PWD
     - $ORIGIN is expanded to the path of the given file, ``origin``

    :param ld_path: Colon-delimited string of paths, defaults to the
        ``LD_LIBRARY_PATH`` environment variable
    :param origin: Directory containing the ELF file being parsed
        (used for $ORIGIN), defaults to an empty string
    :param root: Path to prepend to all paths found
    :return: Iterator over the processed paths
    """

    if ld_path is None:
        ld_path = os.environ.get('LD_LIBRARY_PATH')
        if ld_path is None:
            return
    logger.debug("Parsing ld_path %s", ld_path)

    lib = 'lib64' if platform.architecture()[0] == '64bit' else 'lib'
    platform_ = platform.machine()
    for path in ld_path.split(':'):
        if not path:
            yield normpath(os.getcwd())
        else:
            for k in (('$ORIGIN', origin), ('${ORIGIN}', origin),
                      ('$LIB', lib), ('${LIB}', lib),
                      ('$PLATFORM', platform_), ('${PLATFORM}', platform_)):
                path = path.replace(*k)
            if os.path.isabs(path):
                path = root + '/' + path
            yield normpath(path)


def parse_ld_so_conf_iter(conf_path: Optional[str] = None, root: str = '/') \
        -> Iterator[str]:
    """Parse a ldso config file

    This should handle comments, whitespace, and "include" statements.

    :param conf_path: Path of the ldso config file to parse,
        defaults to ``{root}/etc/ld.so.conf``
    :param root: Path to prepend to all paths found
    :return: Iterator over the processed paths
    """
    if conf_path is None:
        conf_path = normpath(root + '/etc/ld.so.conf')
        if not os.path.isfile(conf_path):
            return
    logger.debug("Parsing ld.so.conf %s", conf_path)

    with open(conf_path, 'r') as conf_file:
        for line in conf_file:
            line = line.split('#', 1)[0].strip()
            if not line:
                continue
            if line.startswith('include '):
                line = line[8:]
                if line[0] != '/':
                    line = os.path.dirname(conf_path) + '/' + line
                for path in sorted(glob.glob(line)):
                    yield from parse_ld_so_conf_iter(normpath(path), root)
            else:
                yield normpath(root + line)


@functools.lru_cache()
def parse_ld_so_conf_tuple(conf_path: Optional[str] = None, root: str = '/') \
        -> Tuple[str, ...]:
    """Parse a ldso config file

    Cached version of :func:`parse_ld_so_conf_iter`, returning a tuple.

    :param conf_path: Path of the ldso config file to parse,
        defaults to ``{root}/etc/ld.so.conf``
    :param root: Path to prepend to all paths found
    :return: Tuple with the processed paths
    """
    return tuple(parse_ld_so_conf_iter(conf_path, root))


@functools.lru_cache()
def _get_default_libdirs(root: str = '/') -> Tuple[str, ...]:
    """Get the default library directories

    :param root: Root directory to check for library directories
    :return: Libdirs in the initramfs
    """
    libdirs = []
    for lib in ('lib64', 'lib', 'lib32'):
        for prefix in ('/', '/usr/'):
            path = normpath(root + prefix + lib)
            if os.path.exists(path):
                libdirs.append(path)
    return tuple(libdirs)


@functools.lru_cache()
def _get_libdir(arch: int, root: str = '/') -> str:
    """Get the libdir corresponding to a binary class

    :param arch: Binary class (e.g. :data:`32` or :data:`64`)
    :param root: Directory where libdirs are searched
    :return: Libdir in the initramfs
    """
    if arch == 64 and os.path.exists(root + '/lib64'):
        return '/lib64'
    if arch == 32 and os.path.exists(root + '/lib32'):
        return '/lib32'
    return '/lib'


def _is_elf_compatible(elf1: ELFFile, elf2: ELFFile) -> bool:
    """See if two ELFs are compatible

    This compares the aspects of the ELF to see if they're compatible:
    bit size, endianness, machine type, and operating system.

    :param elf1: First ELF object
    :param elf2: Second ELF object
    :return: :data:`True` if compatible, :data:`False` otherwise
    """

    osabis = frozenset([e.header['e_ident']['EI_OSABI'] for e in (elf1, elf2)])
    compat_sets = (frozenset(
        'ELFOSABI_%s' % x for x in ('NONE', 'SYSV', 'GNU', 'LINUX',)
    ),)
    return (
        (len(osabis) == 1 or any(osabis.issubset(x) for x in compat_sets))
        and elf1.elfclass == elf2.elfclass
        and elf1.little_endian == elf2.little_endian
        and elf1.header['e_machine'] == elf2.header['e_machine']
    )


def _get_elf_arch(elf1: Union[ELFFile, str], elf2: Union[ELFFile, str]) -> int:
    """Open elf2, check compatibility, and return ELF architecture

    :param elf1: First ELF
    :param elf2: Second ELF
    :return: Architecture of the ELF files (32 or 64)
    :raises OSError: Could not open an ELF file
    :raises ELFIncompatibleError: ELFs are incompatible
    :raises ELFError: File is not an ELF file
    """

    # Convert string to ELFFile
    if isinstance(elf1, str):
        with open(elf1, 'rb') as elf_file:
            return _get_elf_arch(ELFFile(elf_file), elf2)
    if isinstance(elf2, str):
        with open(elf2, 'rb') as elf_file:
            return _get_elf_arch(elf1, ELFFile(elf_file))
    if not _is_elf_compatible(elf1, elf2):
        raise ELFIncompatibleError("Incompatible ELF binaries")
    return elf1.elfclass


def _find_elf_deps_iter(elf: ELFFile, origin: str, root: str = '/') \
        -> Iterator[Tuple[str, str]]:
    """Iterates over the dependencies of an ELF file

    Backend of :func:`find_elf_deps_iter`.

    :param elf: Elf file to parse
    :param origin: Directory containing the ELF binary (real path as provided
        by :func:`os.path.realpath`, used for $ORIGIN)
    :param root: Path to prepend to all paths found
    :return: Same as :func:`find_elf_deps_iter`
    :raises FileNotFoundError: Dependency not found
    """
    deps: List[str] = []
    rpaths: List[str] = []
    runpaths: List[str] = []
    nodeflib = False

    # Read ELF segments
    for segment in elf.iter_segments():
        if segment.header.p_type == 'PT_INTERP':
            interp = segment.get_interp_name()
            logger.debug("INTERP: %s", interp)
            deps.append(interp)
        elif segment.header.p_type == 'PT_DYNAMIC':
            for tag in segment.iter_tags():
                if tag.entry.d_tag == 'DT_RPATH':
                    rpaths.extend(parse_ld_path(tag.rpath, origin, root))
                elif tag.entry.d_tag == 'DT_RUNPATH':
                    runpaths.extend(parse_ld_path(tag.runpath, origin, root))
                elif tag.entry.d_tag == 'DT_NEEDED':
                    deps.extend(parse_ld_path(tag.needed, origin, root))
                elif tag.entry.d_tag == 'DT_FLAGS_1':
                    if tag.entry.d_val & ENUM_DT_FLAGS_1['DF_1_NODEFLIB']:
                        nodeflib = True

    logger.debug("ELF: deps: %s, rpaths: %s, runpaths: %s, nodeflib: %s",
                 deps, rpaths, runpaths, nodeflib)

    # Directories in which dependencies will be searched
    search_paths_base = tuple(itertools.chain(
        rpaths, parse_ld_path(None, origin, root), runpaths,
        parse_ld_so_conf_tuple(root=root), _get_default_libdirs(root)
    ))

    for dep in deps:

        search_paths = search_paths_base \
            if not os.path.isabs(dep) else (root,)

        # Search the dependency
        for found_dir in search_paths:
            found_path = normpath(found_dir + '/' + dep)

            # Check found_path is valid and compatible
            if nodeflib and found_dir in _get_default_libdirs(root):
                continue
            try:
                found_arch = _get_elf_arch(elf, found_path)
            except (ELFError, OSError):
                continue

            # Lib found
            if os.path.isabs(dep) \
                    or found_dir in itertools.chain(rpaths, runpaths):
                # Libdir in R*PATH: use the same path
                dest = normpath('/' + removeprefix(found_path, root))
            else:
                # Libdir in ld_path or ld.so.conf: use default libdir
                dest = normpath(_get_libdir(found_arch, root) + '/' + dep)
            logger.debug("Found %s in %s (dest: %s)", dep, found_dir, dest)
            yield found_path, dest
            break
        else:
            raise FileNotFoundError(f"ELF dependency not found: {dep}")


def find_elf_deps_iter(src: str, root: str = '/') -> Iterator[Tuple[str, str]]:
    """Iterates over the dependencies of an ELF file

    Read an ELF file to search dynamic library dependencies. For each
    dependency, find it on the system (using RPATH, LD_LIBRARY_PATH,
    RUNPATH, ld.so.conf, and default library directories).

    If the library is in a path encoded in the ELF binary (RPATH or RUNPATH),
    ``dep_dest = dep_src``, otherwise use a default library directory
    according to the type of binary (``/lib``, ``/lib64``, ``/lib32``).

    If the file is not an ELF file, returns an empty iterator.

    :param src: File to find dependencies for
    :param root: Path to prepend to all paths found
    :return: Iterator of ``(dep_src, dep_dest)``, with ``dep_src`` the path
        of the dependency on the current system, and ``dep_dest`` the
        path of the dependency on the initramfs
    :raises FileNotFoundError: Dependency not found
    """
    logger.debug("Searching ELF dependencies for %s", src)

    if src != os.path.realpath(src):
        yield from find_elf_deps_iter(os.path.realpath(src), root)
        return

    with open(src, 'rb') as src_file:
        try:
            elf = ELFFile(src_file)
        except ELFError:
            return
        yield from _find_elf_deps_iter(elf, os.path.dirname(src), root)
    logger.debug("Found all ELF dependencies for %s", src)


@functools.lru_cache()
def find_elf_deps_set(src: str, root: str = '/') -> FrozenSet[Tuple[str, str]]:
    """Find dependencies of an ELF file

    Cached version of :func:`find_elf_deps_iter`.

    :param src: File to find dependencies for
    :param root: Path to prepend to all paths found
    :return: Set of ``(dep_src, dep_dest)``, see :func:`find_elf_deps_iter`.
    :raises FileNotFoundError: Dependency not found
    """
    if src != os.path.realpath(src):
        return find_elf_deps_set(os.path.realpath(src), root)
    return frozenset(find_elf_deps_iter(src, root))


def find_lib_iter(lib: str, compat: Optional[str] = None, root: str = '/') \
        -> Iterator[Tuple[str, str]]:
    """Search a library in the system, with globbing

    Same as :func:`find_lib` but uses :func:`glob.glob` to find matching
    libraries.

    :param lib: Glob pattern for the library to search (e.g. ``libgcc_s.*``)
    :param compat: Path to a binary that the library must be compatible with
        (checked with :func:`_is_elf_compatible`),
        defaults to ``{root}/bin/sh``
    :param root: Path to prepend to all paths found
    :return: Iterator over ``(lib_src, lib_dest)``, see :func:`find_lib`
    :raises FileNotFoundError: Library not found
    """
    if compat is None:
        compat = normpath(root + '/bin/sh')
    logger.debug("Searching library %s (compat: %s)", lib, compat)

    libname = os.path.basename(lib)
    with open(compat, 'rb') as pyexec:
        pyelf = ELFFile(pyexec)

        # If path is absolute: only search in root
        search_paths = itertools.chain(
            (os.getcwd(),),
            parse_ld_path(root=root),
            parse_ld_so_conf_tuple(root=root),
            _get_default_libdirs(root)
        ) if not os.path.isabs(lib) else (root,)

        found = False
        for found_dir in search_paths:
            found_path = found_dir + '/' + libname

            for found_path in glob.iglob(found_path):
                try:
                    found_arch = _get_elf_arch(pyelf, found_path)
                except (ELFError, OSError):
                    continue
                found = True
                dest = normpath(_get_libdir(found_arch, root) + '/' + libname)
                logger.debug("Found %s in %s (dest: %s)", lib, found_dir, dest)
                yield found_path, dest

    if not found:
        raise FileNotFoundError(lib)


def find_lib(lib: str, compat: Optional[str] = None, root: str = '/') \
        -> Tuple[str, str]:
    """Search a library in the system, without globbing

    Uses ``ld.so.conf`` and ``LD_LIBRARY_PATH``.

    Libraries will be installed in the default library directory in the
    initramfs.

    :param lib: Library to search (e.g. ``libgcc_s.so.1``)
    :param compat: Path to a binary that the library must be compatible with
        (checked with :func:`_is_elf_compatible`),
        defaults to ``{root}/bin/sh``
    :param root: Path to prepend to all paths found
    :return: ``(lib_src, lib_dest)``, with ``lib_src`` the absolute path
        of the library on the current system, and ``lib_dest`` the absolute
        path of the library on the initramfs
    :raises FileNotFoundError: Library not found
    """
    return next(find_lib_iter(glob.escape(lib), compat, root))


def parse_path(path: Optional[str] = None, root: str = '/') \
        -> Iterator[str]:
    """Parse PATH variable

    :param path: PATH string to parse,
        default to the ``PATH`` environment variable
    :param root: Path to prepend to all paths found
    :return: Iterator over the processed paths
    """
    if path is None:
        path = os.environ.get('PATH')
        if path is None:
            return
    logger.debug("Parsing path %s", path)

    for k in path.split(':'):
        if not k:
            yield normpath(os.getcwd())
        else:
            yield normpath(root + '/' + k)


def find_exec(executable: str, compat: Optional[str] = None, root: str = '/') \
        -> Tuple[str, str]:
    """Search an executable in the system

    Uses the ``PATH`` environment variable.

    :param executable: Executable to search
    :param compat: Path to a binary that the executable must be compatible with
        (checked with :func:`_is_elf_compatible`),
        defaults to ``{root}/bin/sh``
    :param root: Path to prepend to all paths found
    :return: ``(src_path, dest_path)`` with ``src_path`` the path of the
        executable on ``root``, and ``dest_path`` the default path of the
        executable on the initramfs.
    :return: Absolute path of the executable
    :raises FileNotFoundError: Executable not found
    """
    if compat is None:
        compat = normpath(root + '/bin/sh')
    logger.debug("Searching executable %s (compat: %s)", executable, compat)

    # Get list of directories to search
    execdirs = itertools.chain((os.getcwd(),), parse_path(root=root)) \
        if not os.path.isabs(executable) else (root,)

    # Parse directories
    execname = os.path.basename(executable)
    with open(compat, 'rb') as compat_file:
        compat_elf = ELFFile(compat_file)

        for found_dir in execdirs:
            found_path = normpath(found_dir + '/' + execname)

            # Check for compatibility
            if not os.access(found_path, mode=os.X_OK):
                continue
            try:
                _get_elf_arch(compat_elf, found_path)
            except (ELFIncompatibleError, OSError):
                continue
            except ELFError:
                pass
            dest = normpath('/' + removeprefix(found_path, root))
            logger.debug("Found %s in %s (dest: %s)",
                         executable, found_dir, dest)
            return found_path, dest
    raise FileNotFoundError(executable)


@functools.lru_cache()
def _get_all_kmods(kernel: str) -> FrozenSet[str]:
    """Get all kernel modules on the system

    :param kernel: Target kernel version
    :return: Set with the absolute path of the modules
    """
    return frozenset(glob.glob(normpath(f'{KMOD_DIR}/{kernel}/**/*.ko')))


@functools.lru_cache()
def find_kmod_deps(path: str) -> FrozenSet[str]:
    """Get kernel module dependencies

    :param path: Path of the kernel module to parse
    :return: Set with the dependencies' names
    :raises subprocess.CalledProcessError: Error during ``modinfo``
    """

    cmd = ('modinfo', '-0', '-F', 'depends', path)
    logger.debug("Subprocess: %s", cmd)
    proc = subprocess.run(cmd, stdout=subprocess.PIPE, check=True)
    return frozenset(k for k in proc.stdout.decode('UTF-8').split('\0') if k)


def find_kmod(module: str, kernel: str) -> str:
    """Search a kernel module on the system

    :param module: Name of the kernel module
    :param kernel: Target kernel version
    :return: Absolute path of the kernel module on the system
    :raises FileNotFoundError: Kernel module not found
    """

    logger.debug("Searching module %s for kernel %s", module, kernel)
    if os.path.isabs(module):
        logger.debug("Module path is absolute: %s", module)
        return module
    module_compat = module.replace('_', '-') + '.ko'
    for kmod in _get_all_kmods(kernel):
        if module_compat == os.path.basename(kmod).replace('_', '-'):
            kmod = normpath(kmod)
            logger.debug("Found module %s: %s", module, kmod)
            return kmod
    raise FileNotFoundError(f"Kernel module not found: {module}")



"""Module providing the Item class for files in the initramfs

Each file type of the initramfs has an :class:`Item` subclass. Those classes
provide methods used by :class:`Initramfs` generation methods.
"""

from __future__ import annotations

import logging
import os
import socket
import stat
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Iterator, Set

from .utils import hash_file


logger = logging.getLogger(__name__)


class MergeError(Exception):
    """Cannot merge an Item into another"""


@dataclass  # type: ignore
class Item(ABC):
    """An object within the initramfs"""

    def is_mergeable(self, other: Item) -> bool:
        """Check if two items can be merged together

        By default, two items can only be merged if they are equal.

        :param other: :class:`Item` to merge into ``self``
        :return: :data:`True` if the items can be merged,
            :data:`False` otherwise
        """
        return self == other

    def merge(self, other: Item) -> None:
        """Merge two items together

        Default merge is just a no-op. Subclasses can override this
        as done by :class:`File` to handle hardlink of identical files.

        :param other: :class:`Item` to merge into ``self``
        :raises MergeError: Cannot merge the items
        """
        if not self.is_mergeable(other):
            raise MergeError(f"Different items: {self} != {other}")

    def __iter__(self) -> Iterator[str]:
        """Get the paths of this item within the initramfs

        This method should be overriden by subclasses which add files
        to the initramfs.

        :return: Iterator over this :class:`Item`'s destination paths
        """
        return iter(())

    def __contains__(self, path: str) -> bool:
        """Check if this item is present at the given path in the initramfs

        This method should be overriden by subclasses which add files
        to the initramfs.

        :return: :data:`True` if ``path`` is part of this :class:`Item`'s
            destination paths, :data:`False` otherwise
        """
        return False

    @staticmethod
    def build_from_cpio_list(data: str) -> Item:
        """Build an Item from a string

        This string should respect the format of ``gen_init_cpio``.

        :param data: String to parse
        :return: Item corresponding to ``data``
        :raises ValueError: Invalid string
        """
        parts = data.split()
        if parts[0] == 'file' and len(parts) >= 6:
            return File(
                int(parts[3], base=8), int(parts[4]), int(parts[5]),
                set(parts[6:] + [parts[1]]), parts[2], hash_file(parts[2])
            )
        if parts[0] == 'dir' and len(parts) == 5:
            return Directory(
                int(parts[2], base=8), int(parts[3]), int(parts[4]), parts[1]
            )
        if parts[0] == 'nod' and len(parts) == 8:
            return Node(
                int(parts[2], base=8), int(parts[3]), int(parts[4]), parts[1],
                Node.NodeType(parts[5]), int(parts[6]), int(parts[7])
            )
        if parts[0] == 'slink' and len(parts) == 6:
            return Symlink(
                int(parts[3], base=8), int(parts[4]), int(parts[5]),
                parts[1], parts[2]
            )
        if parts[0] == 'pipe' and len(parts) == 5:
            return Pipe(
                int(parts[2], base=8), int(parts[3]), int(parts[4]), parts[1]
            )
        if parts[0] == 'sock' and len(parts) == 5:
            return Socket(
                int(parts[2], base=8), int(parts[3]), int(parts[4]), parts[1]
            )
        if parts[0] in ('file', 'dir', 'nod', 'slink', 'pipe', 'sock'):
            raise ValueError(f"Invalid format for {parts[0]}: {parts[1:]}")
        raise ValueError(f"Unknown type: {parts[0]}")

    @abstractmethod
    def build_to_cpio_list(self) -> str:
        """String representing the item

        The string is formatted to be compatible with the ``gen_init_cpio``
        tool from the Linux kernel.
        This method has to be defined by subclasses.
        """

    @abstractmethod
    def build_to_directory(self, base_dir: str) -> None:
        """Add this item to a real filesystem

        This will copy or create a file on a real filesystem.
        This method has to be defined by subclasses.

        :param base_dir: Path to use as root directory (e.g. using
            ``/tmp/initramfs``, ``/bin/ls`` will be copied to
            ``/tmp/initramfs/bin/ls``)
        """


@dataclass
class File(Item):
    """Normal file within the initramfs

    :param mode: Permissions (e.g. 0o644)
    :param user: Owner user (UID)
    :param group: Owner group (GID)
    :param dests: Paths in the initramfs (hard-linked)
    :param src: Source file to copy (not unique to the file)
    :param data_hash: Hash of the file (can be obtained with :func:`hash_file`)
    :param chunk_size: Chunk size to use when copying the file
    """
    mode: int
    user: int
    group: int
    dests: Set[str]
    src: str
    data_hash: bytes
    chunk_size: int = 65536

    def __str__(self) -> str:
        return f"file from {self.src}"

    def is_mergeable(self, other: Item) -> bool:
        return isinstance(other, File) \
            and self.data_hash == other.data_hash \
            and self.mode == other.mode \
            and self.user == other.user \
            and self.group == other.group

    def merge(self, other: Item) -> None:
        if self.is_mergeable(other):
            assert isinstance(other, File)
            self.dests |= other.dests
        else:
            raise MergeError(f"Different files: {self} and {other}")

    def __iter__(self) -> Iterator[str]:
        return iter(self.dests)

    def __contains__(self, path: str) -> bool:
        return path in self.dests

    def build_to_cpio_list(self) -> str:
        dests = iter(sorted(self.dests))
        return f'file {next(dests)} {self.src} ' \
            + f'{self.mode:03o} {self.user} {self.group}' \
            + (' ' if len(self.dests) > 1 else '') \
            + ' '.join(dests)

    def build_to_directory(self, base_dir: str) -> None:
        iter_dests = iter(self.dests)
        # Copy reference file
        base_dest = base_dir + next(iter_dests)
        with open(self.src, 'rb') as src_file, \
                open(base_dest, 'wb') as dest_file:
            for chunk in iter(lambda: src_file.read(self.chunk_size), b''):
                dest_file.write(chunk)
        os.chmod(base_dest, self.mode)
        os.chown(base_dest, self.user, self.group)
        # Hardlink other files
        for dest in iter_dests:
            abs_dest = base_dir + dest
            os.link(base_dest, abs_dest)


@dataclass
class Directory(Item):
    """Directory within the initramfs

    :param mode: Permissions (e.g. 0o644)
    :param user: Owner user (UID)
    :param group: Owner group (GID)
    :param dest: Path in the initramfs
    """
    mode: int
    user: int
    group: int
    dest: str

    def __str__(self) -> str:
        return f"directory {self.dest}"

    def __iter__(self) -> Iterator[str]:
        return iter((self.dest,))

    def __contains__(self, path: str) -> bool:
        return path == self.dest

    def build_to_cpio_list(self) -> str:
        return f'dir {self.dest} {self.mode:03o} {self.user} {self.group}'

    def build_to_directory(self, base_dir: str) -> None:
        abs_dest = base_dir + self.dest
        os.mkdir(abs_dest)
        os.chmod(abs_dest, self.mode)
        os.chown(abs_dest, self.user, self.group)


@dataclass
class Node(Item):
    """Special file within the initramfs

    :param mode: Permissions (e.g. 0o644)
    :param user: Owner user (UID)
    :param group: Owner group (GID)
    :param dest: Path in the initramfs
    :param nodetype: Type of node (block, character)
    :param major: Major number of the node
    :param minor: Minor number of the node
    """
    mode: int
    user: int
    group: int
    dest: str
    nodetype: Node.NodeType
    major: int
    minor: int

    class NodeType(Enum):
        """Special file type"""
        #: Block device
        BLOCK = 'b'
        #: Character device
        CHARACTER = 'c'

    def __str__(self) -> str:
        if self.nodetype == Node.NodeType.CHARACTER:
            return f"character device {self.major} {self.minor} {self.dest}"
        if self.nodetype == Node.NodeType.BLOCK:
            return f"block device {self.major} {self.minor} {self.dest}"
        raise ValueError(f"Unknown node type {self.nodetype}")

    def __iter__(self) -> Iterator[str]:
        return iter((self.dest,))

    def __contains__(self, path: str) -> bool:
        return path == self.dest

    def build_to_cpio_list(self) -> str:
        return f'nod {self.dest} {self.mode:03o} {self.user} {self.group} ' \
            f'{self.nodetype.value} {self.major} {self.minor}'

    def build_to_directory(self, base_dir: str) -> None:
        abs_dest = base_dir + self.dest
        if self.nodetype == Node.NodeType.BLOCK:
            mode = self.mode | stat.S_IFBLK
        elif self.nodetype == Node.NodeType.CHARACTER:
            mode = self.mode | stat.S_IFCHR
        os.mknod(abs_dest, mode, os.makedev(self.major, self.minor))
        os.chmod(abs_dest, self.mode)
        os.chown(abs_dest, self.user, self.group)


@dataclass
class Symlink(Item):
    """Symlinks within the initramfs

    :param mode: Permissions (e.g. 0o644)
    :param user: Owner user (UID)
    :param group: Owner group (GID)
    :param dest: Path in the initramfs
    :param target: Link target
    """
    mode: int
    user: int
    group: int
    dest: str
    target: str

    def __str__(self) -> str:
        return f"symlink {self.dest} to {self.target}"

    def __iter__(self) -> Iterator[str]:
        return iter((self.dest,))

    def __contains__(self, path: str) -> bool:
        return path == self.dest

    def build_to_cpio_list(self) -> str:
        return f'slink {self.dest} {self.target} ' \
            f'{self.mode:03o} {self.user} {self.group}'

    def build_to_directory(self, base_dir: str) -> None:
        if self.mode != 0o777:
            logger.warning("Cannot set mode for %s", self)
        abs_dest = base_dir + self.dest
        os.symlink(self.target, abs_dest)
        os.chown(abs_dest, self.user, self.group, follow_symlinks=False)


@dataclass
class Pipe(Item):
    """Named pipe (FIFO) within the initramfs

    :param mode: Permissions (e.g. 0o644)
    :param user: Owner user (UID)
    :param group: Owner group (GID)
    :param dest: Path in the initramfs
    """
    mode: int
    user: int
    group: int
    dest: str

    def __str__(self) -> str:
        return f"named pipe {self.dest}"

    def __iter__(self) -> Iterator[str]:
        return iter((self.dest,))

    def __contains__(self, path: str) -> bool:
        return path == self.dest

    def build_to_cpio_list(self) -> str:
        return f'pipe {self.dest} {self.mode:03o} {self.user} {self.group}'

    def build_to_directory(self, base_dir: str) -> None:
        abs_dest = base_dir + self.dest
        os.mkfifo(abs_dest)
        os.chmod(abs_dest, self.mode)
        os.chown(abs_dest, self.user, self.group)


@dataclass
class Socket(Item):
    """Named socket within the initramfs

    :param mode: Permissions (e.g. 0o644)
    :param user: Owner user (UID)
    :param group: Owner group (GID)
    :param dest: Path in the initramfs
    """
    mode: int
    user: int
    group: int
    dest: str

    def __str__(self) -> str:
        return f"named socket {self.dest}"

    def __iter__(self) -> Iterator[str]:
        return iter((self.dest,))

    def __contains__(self, path: str) -> bool:
        return path == self.dest

    def build_to_cpio_list(self) -> str:
        return f'sock {self.dest} {self.mode:03o} {self.user} {self.group}'

    def build_to_directory(self, base_dir: str) -> None:
        abs_dest = base_dir + self.dest
        sock = socket.socket(socket.AF_UNIX)
        sock.bind(abs_dest)
        os.chmod(abs_dest, self.mode)
        os.chown(abs_dest, self.user, self.group)



"""Entry point module for cmkinitramfs"""

from __future__ import annotations

import argparse
import configparser
import itertools
import locale
import logging
import os
import os.path
import shlex
import shutil
import sys
from dataclasses import dataclass
from typing import Dict, Iterable, List, Mapping, Optional, Tuple, overload

import cmkinitramfs
import cmkinitramfs.data as datamod
import cmkinitramfs.initramfs as mkramfs
from .bin import find_lib, find_lib_iter
from .init import (mkinit, Breakpoint, BUSYBOX_COMMON_DEPS,
                   BUSYBOX_KEYMAP_DEPS, BUSYBOX_KMOD_DEPS)
from .utils import removeprefix

logger = logging.getLogger(__name__)
_VERSION_INFO = \
    f"%(prog)s ({cmkinitramfs.__name__}) {cmkinitramfs.__version__}"
BINARY_KEYMAP_MAGIC = b'bkeymap'


def _find_config_file() -> str:
    """Find a configuration file to use"""
    env_config = os.environ.get('CMKINITCFG')
    if env_config is not None and os.path.isfile(env_config):
        return env_config
    if os.path.isfile('./cmkinitramfs.ini'):
        return './cmkinitramfs.ini'
    if os.path.isfile('/etc/cmkinitramfs.ini'):
        return '/etc/cmkinitramfs.ini'
    raise FileNotFoundError("Configuration file not found")


@dataclass
class Config:
    """Configuration informations

    :param root: Rootfs data needed to boot
    :param mounts: Non-rootfs datas needed to boot
    :param keymap: Keymap information tuple ``(source, build, dest)``:
        ``source``: keymap to convert, ``build``: converted keymap,
        ``dest``: keymap path within the initramfs
    :param files: User configured files,
        see :attr:`cmkinitramfs.init.Data.files`
    :param execs: User configured executables,
        see :attr:`cmkinitramfs.init.Data.files`
    :param libs: User configured libraries,
        see :attr:`cmkinitramfs.init.Data.files`
    :param busybox: Needed executables compatibles with busybox implementation
    :param init_path: Path where the init script will be generated
    :param cmkcpiodir_opts: Default options for cmkcpiodir
    :param cmkcpiolist_opts: Default options for cmkcpiolist
    :param modules: Kernel modules to be loaded in the initramfs:
        ``{module: (arg, ...)}``. See :func:`cmkinitramfs.init.mkinit`.
    :param scripts: User scripts to run at given breakpoints.
        See ``scripts`` for :func:`cmkinitramfs.init.mkinit`.
    """
    root: datamod.Data
    mounts: Iterable[datamod.Data]
    keymap: Optional[Tuple[str, str, str]]
    files: Iterable[Tuple[str, Optional[str]]]
    execs: Iterable[Tuple[str, Optional[str]]]
    libs: Iterable[Tuple[str, Optional[str]]]
    busybox: Iterable[str]
    init_path: str
    cmkcpiodir_opts: str
    cmkcpiolist_opts: str
    modules: Mapping[str, Iterable[str]]
    scripts: Mapping[Breakpoint, Iterable[str]]


def read_config(config_file: Optional[str] = None) -> Config:
    """Read a configuration file and generate data structures from it

    :param config_file: Configuration file to use. Defaults to, in order:
        ``CMKINITCFG`` environment variable, ``./cmkinitramfs.ini``,
        ``/etc/cmkinitramfs.ini``.
    :return: Configuration dictionnary, described by :class:`Config`
    :raises ValueError: Config file parsing error
    """

    @overload
    def find_data(data_str: None) -> None: ...
    @overload
    def find_data(data_str: str) -> datamod.Data: ...

    def find_data(data_str: Optional[str]) -> Optional[datamod.Data]:
        """Find a Data object from a data string"""
        if data_str is None:
            return None
        if data_str.startswith('PATH='):
            data_str = removeprefix(data_str, 'PATH=')
            if data_dic.get(data_str) is None:
                data_dic[data_str] = datamod.PathData(data_str)
        elif data_str.startswith('UUID='):
            data_str = removeprefix(data_str, 'UUID=')
            if data_dic.get(data_str) is None:
                data_dic[data_str] = datamod.UuidData(data_str, False)
        elif data_str.startswith('LABEL='):
            data_str = removeprefix(data_str, 'LABEL=')
            if data_dic.get(data_str) is None:
                data_dic[data_str] = datamod.LabelData(data_str, False)
        elif data_str.startswith('PARTUUID='):
            data_str = removeprefix(data_str, 'PARTUUID=')
            if data_dic.get(data_str) is None:
                data_dic[data_str] = datamod.UuidData(data_str, True)
        elif data_str.startswith('PARTLABEL='):
            data_str = removeprefix(data_str, 'PARTLABEL=')
            if data_dic.get(data_str) is None:
                data_dic[data_str] = datamod.LabelData(data_str, True)
        elif data_str.startswith('DATA='):
            data_str = removeprefix(data_str, 'DATA=')
        elif data_dic.get(data_str) is None and os.path.isabs(data_str):
            data_dic[data_str] = datamod.PathData(data_str)
        return data_dic[data_str]

    # Read config file
    if config_file is None:
        config_file = _find_config_file()
    config = configparser.ConfigParser()
    if config.read(config_file) != [config_file]:
        raise ValueError(f"Cound not read configuration {config_file}")

    # Get all data sources in data_dic
    data_dic: Dict[str, datamod.Data] = {}
    for data_id in config.sections():
        data_config = config[data_id]
        if data_config['type'] == 'luks':
            data_dic[data_id] = datamod.LuksData(
                find_data(data_config['source']),
                data_config['name'],
                find_data(data_config.get('key')),
                find_data(data_config.get('header')),
                data_config.getboolean('discard', fallback=False),
            )
        elif data_config['type'] == 'lvm':
            data_dic[data_id] = datamod.LvmData(
                data_config['vg-name'],
                data_config['lv-name'],
            )
        elif data_config['type'] == 'mount':
            data_dic[data_id] = datamod.MountData(
                find_data(data_config['source']),
                data_config['mountpoint'],
                data_config['filesystem'],
                data_config.get('options', 'ro'),
            )
        elif data_config['type'] == 'md':
            data_dic[data_id] = datamod.MdData(
                [find_data(k.strip())
                 for k in data_config['source'].strip().split('\n')],
                data_config['name'],
            )
        else:
            raise Exception(f"Unknown config type {data_config['type']}")

    # Configure dependencies
    for data_id, data in data_dic.items():
        if data_id not in config.sections():
            continue
        data_config = config[data_id]
        for dep in data_config.get('need', '').split(','):
            if dep.strip():
                data.add_dep(find_data(dep.strip()))
        for ldep in data_config.get('load-need', '').split(','):
            if ldep.strip():
                data.add_load_dep(find_data(ldep.strip()))

    # Define Data for root and for other mounts
    root = find_data(config['DEFAULT']['root'])
    mounts = tuple(
        find_data(k.strip())
        for k in config['DEFAULT'].get('mountpoints', '').split(',')
        if k.strip()
    )

    # Define needed files, execs and libs
    files = set()
    for data in itertools.chain((root,), mounts):
        files |= data.files
        for ddep in data.iter_all_deps():
            files |= ddep.files
    for line in config['DEFAULT'].get('files', '').split('\n'):
        if line:
            src, *dest = line.split(':', maxsplit=1)
            files.add((src, dest[0] if dest else None))
    execs = set()
    for data in itertools.chain((root,), mounts):
        execs |= data.execs
        for ddep in data.iter_all_deps():
            execs |= ddep.execs
    for line in config['DEFAULT'].get('execs', '').split('\n'):
        if line:
            src, *dest = line.split(':', maxsplit=1)
            execs.add((src, dest[0] if dest else None))
    libs = set()
    for data in itertools.chain((root,), mounts):
        libs |= data.libs
        for ddep in data.iter_all_deps():
            libs |= ddep.libs
    for line in config['DEFAULT'].get('libs', '').split('\n'):
        if line:
            src, *dest = line.split(':', maxsplit=1)
            libs.add((src, dest[0] if dest else None))
    busybox = set()
    for data in itertools.chain((root,), mounts):
        busybox |= data.busybox
        for ddep in data.iter_all_deps():
            busybox |= ddep.busybox
    for line in config['DEFAULT'].get('busybox', '').split('\n'):
        if line:
            busybox.add(line.strip())

    modules: Dict[str, List[str]] = {}
    for module in config['DEFAULT'].get('modules', '').split('\n'):
        if module:
            mod_name, *mod_args = module.split()
            if modules.get(mod_name) is None:
                modules[mod_name] = []
            modules[mod_name].extend(mod_args)

    # User scripts
    breakpoints = {
        'early': Breakpoint.EARLY,
        'init': Breakpoint.INIT,
        'module': Breakpoint.MODULE,
        'rootfs': Breakpoint.ROOTFS,
        'mount': Breakpoint.MOUNT,
    }
    scripts: Dict[Breakpoint, List[str]] = {k: [] for k in Breakpoint}
    for script in config['DEFAULT'].get('scripts', '').split('\n'):
        if script:
            bname, script = script.split(':', maxsplit=1)
            scripts[breakpoints[bname.strip().lower()]].append(script.strip())

    # Create dictionnary to return
    ret_cfg = Config(
        root=root,
        mounts=mounts,
        keymap=(
            config['DEFAULT'].get('keymap-src'),
            config['DEFAULT'].get('keymap-path', '/tmp/keymap.bmap'),
            config['DEFAULT'].get('keymap-dest', '/root/keymap.bmap'),
        ) if config['DEFAULT'].getboolean('keymap', fallback=False) else None,
        files=files,
        execs=execs,
        libs=libs,
        busybox=busybox,
        init_path=config['DEFAULT'].get('init-path', '/tmp/init.sh'),
        cmkcpiodir_opts=config['DEFAULT'].get(
            'cmkcpiodir-default-opts', ''
        ),
        cmkcpiolist_opts=config['DEFAULT'].get(
            'cmkcpiolist-default-opts', ''
        ),
        modules=modules,
        scripts=scripts,
    )

    # Configure final data sources
    for data in itertools.chain(ret_cfg.mounts, (ret_cfg.root,)):
        data.set_final()

    return ret_cfg


def entry_cmkinit() -> None:
    """Main entry point of the module"""
    config = read_config()
    parser = argparse.ArgumentParser(description="Build an init script")
    parser.add_argument('--version', action='version', version=_VERSION_INFO)
    parser.parse_args()
    mkinit(
        out=sys.stdout,
        root=config.root,
        mounts=config.mounts,
        keymap=(None if config.keymap is None else config.keymap[2]),
        modules=config.modules,
        scripts=config.scripts,
    )


def _common_parser_logging(verbose: bool = False, quiet: int = 0) \
        -> argparse.ArgumentParser:
    """Create the common parser for entry points with a logger

    :param verbose: Default verbose value
    :param quiet: Default quiet value
    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        '--verbose', '-v', action='store_true', default=verbose,
        help="be verbose",
    )
    parser.add_argument(
        '--quiet', '-q', action='count', default=quiet,
        help="be quiet (can be repeated)",
    )
    return parser


def _set_logging_level(verbose: bool, quiet: int) -> None:
    """Set global logging level according to verbose and quiet"""
    if verbose:
        level = logging.DEBUG
    elif quiet >= 3:
        level = logging.CRITICAL
    elif quiet >= 2:
        level = logging.ERROR
    elif quiet >= 1:
        level = logging.WARNING
    else:
        level = logging.INFO
    logging.getLogger().setLevel(level)


def entry_findlib() -> None:
    """Entry point for the findlib utility"""

    parser = argparse.ArgumentParser(
        description="Find a library on the system",
        parents=(_common_parser_logging(),),
    )
    parser.add_argument('--version', action='version', version=_VERSION_INFO)
    parser.add_argument(
        '--compatible', '-c', type=str, default=None,
        help="set a binary the library must be compatible with",
    )
    parser.add_argument(
        '--root', '-r', type=str, default='/',
        help="set the root directory to search for the library",
    )
    parser.add_argument(
        '--null', '-0', action='store_true', default=False,
        help="paths will be delemited by null characters instead of newlines",
    )
    parser.add_argument(
        '--glob', '-g', action='store_true', default=False,
        help="library names are glob patterns",
    )
    parser.add_argument(
        'libs', metavar='LIB', type=str, nargs='+',
        help="library to search",
    )
    args = parser.parse_args()
    _set_logging_level(args.verbose, args.quiet + 1)

    errors = False
    for lib in args.libs:
        logger.info("Searching library: %s", lib)

        try:
            lib_iter = \
                (find_lib(lib, compat=args.compatible, root=args.root),) \
                if not args.glob \
                else find_lib_iter(lib, compat=args.compatible, root=args.root)
            for found, _ in lib_iter:
                if args.quiet < 3:
                    print(found, end=('\n' if not args.null else '\0'))
        except FileNotFoundError:
            logger.error("%s: Library not found", lib)
            errors = True
            continue
    sys.exit(0 if not errors else 1)


def _common_parser_cmkcpio() -> argparse.ArgumentParser:
    """Create the common parser for cmkcpio* entry points"""
    parser = argparse.ArgumentParser(
        add_help=False,
        parents=(_common_parser_logging(),),
    )
    parser.add_argument(
        '--version', action='version', version=_VERSION_INFO
    )
    parser.add_argument(
        "--debug", "-d", action="store_true", default=False,
        help="debugging mode: non-root, implies -k"
    )
    parser.add_argument(
        "--output", "-o", type=str, default='/usr/src/initramfs.cpio',
        help="set the output of the CPIO archive"
    )
    parser.add_argument(
        '--binroot', '-r', type=str, default='/',
        help="set the root directory for binaries (executables and libraries)"
    )
    parser.add_argument(
        '--kernel', '-K', action='append', type=str, default=None,
        help=("set the target kernel versions of the initramfs, "
              "defaults to the running kernel")
    )
    parser.add_argument(
        '--no-kmod', action='store_true', default=False,
        help="disable kernel modules support",
    )
    return parser


def _build_initramfs(initramfs: mkramfs.Initramfs, config: Config) -> None:
    """Add files to the initramfs from the configuration"""
    busybox_deps = set(config.busybox) | BUSYBOX_COMMON_DEPS

    # Add necessary files
    for src, dest in config.files:
        logger.info("Adding file %s", src)
        initramfs.add_file(src, dest)
    for src, dest in config.libs:
        logger.info("Adding library %s", src)
        initramfs.add_library(src, dest)
    for src, dest in config.execs:
        logger.info("Adding executable %s", src)
        initramfs.add_executable(src, dest)

    # Add keymap
    if config.keymap is not None:
        busybox_deps |= BUSYBOX_KEYMAP_DEPS
        logger.info("Adding keymap as %s", config.keymap[2])
        with open(config.keymap[1], 'rb') as bkeymap:
            if bkeymap.read(len(BINARY_KEYMAP_MAGIC)) != BINARY_KEYMAP_MAGIC:
                logger.error("Binary keymap %s: bad file format",
                             config.keymap[1])
        initramfs.add_file(*config.keymap[1:3], mode=0o644)

    # Add module
    if initramfs.kernels:
        busybox_deps |= BUSYBOX_KMOD_DEPS
        for module in config.modules:
            logger.info("Adding kernel module %s", module)
            initramfs.add_kmod(module)

    # Add /init
    logger.info("Adding init script")
    initramfs.add_file(config.init_path, '/init', mode=0o755)

    # Add busybox
    logger.info("Adding busybox")
    initramfs.add_busybox(needed=busybox_deps)


def entry_cmkcpiolist() -> None:
    """Entry point for cmkcpiolist"""

    # Load configuration
    config = read_config()

    # Arguments
    parser = argparse.ArgumentParser(
        description="Build an initramfs using a CPIO list",
        parents=(_common_parser_cmkcpio(),)
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--only-build-archive', '-c', action="store_true", default=False,
        help="only build the CPIO archive from an existing CPIO list"
    )
    group.add_argument(
        '--only-build-list', '-L', action="store_true", default=False,
        help="only build the CPIO list, implies -k"
    )
    parser.add_argument(
        '--keep', '-k', action="store_true", default=False,
        help="keep the created CPIO list"
    )
    parser.add_argument(
        '--cpio-list', '-l', type=str, default='/tmp/initramfs.list',
        help="set the location of the CPIO list"
    )
    args = parser.parse_args(
        shlex.split(config.cmkcpiolist_opts, posix=True) + sys.argv[1:]
    )

    _set_logging_level(args.verbose, args.quiet)

    # Parse arguments
    if args.debug or args.only_build_list:
        args.keep = True

    # Keymap
    if config.keymap is not None and config.keymap[0] is not None:
        with open(config.keymap[1], 'wb') as keymap_bin:
            mkramfs.keymap_build(
                config.keymap[0], keymap_bin,
                unicode=(locale.getdefaultlocale()[1] == 'UTF-8')
            )

    # Init
    with open(config.init_path, 'w') as init_file:
        mkinit(
            out=init_file,
            root=config.root,
            mounts=config.mounts,
            keymap=(None if config.keymap is None else config.keymap[2]),
            modules=config.modules,
            scripts=config.scripts,
        )

    # Initramfs
    if not args.only_build_archive:
        assert config.init_path is not None
        logger.info("Creating initramfs")
        initramfs = mkramfs.Initramfs(
            user=(0 if not args.debug else os.getuid()),
            group=(0 if not args.debug else os.getgid()),
            binroot=args.binroot,
            kernels=(() if args.no_kmod else args.kernel),
        )
        _build_initramfs(initramfs, config)

        # CPIO list
        logger.info("Generating CPIO list")
        if args.cpio_list == '-' and args.only_build_list:
            initramfs.build_to_cpio_list(sys.stdout)
        else:
            with open(args.cpio_list, 'w') as cpiolist:
                initramfs.build_to_cpio_list(cpiolist)

    if not args.only_build_list:
        # Build CPIO archive
        logger.info("Generating CPIO archive to %s", args.output)
        if args.output == '-':
            mkramfs.mkcpio_from_list(args.cpio_list, sys.stdout.buffer)
        else:
            with open(args.output, 'wb') as cpiodest:
                mkramfs.mkcpio_from_list(args.cpio_list, cpiodest)

    if not args.keep:
        # Cleanup temporary files
        if os.path.exists(args.cpio_list):
            logger.info("Cleaning %s", args.cpio_list)
            os.remove(args.cpio_list)
        if config.keymap is not None and os.path.exists(config.keymap[1]):
            logger.info("Cleaning %s", config.keymap[1])
            os.remove(config.keymap[1])
        if os.path.exists(config.init_path):
            logger.info("Cleaning %s", config.init_path)
            os.remove(config.init_path)


def entry_cmkcpiodir() -> None:
    """Entry point for cmkcpiodir"""

    # Load configuration
    config = read_config()

    # Arguments
    parser = argparse.ArgumentParser(
        description="Build an initramfs using a directory.",
        parents=(_common_parser_cmkcpio(),)
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--only-build-archive', '-c', action="store_true", default=False,
        help="only build the CPIO archive from an existing initramfs directory"
    )
    group.add_argument(
        '--only-build-directory', '-D', action="store_true", default=False,
        help="only build the initramfs directory, implies -k"
    )
    parser.add_argument(
        '--keep', '-k', action="store_true", default=False,
        help="keep the created initramfs directory"
    )
    parser.add_argument(
        "--clean", "-C", action="store_true", default=False,
        help="overwrite temporary directory if it exists, use carefully"
    )
    parser.add_argument(
        '--build-dir', '-b', type=str, default='/tmp/initramfs',
        help="set the location of the initramfs directory"
    )
    args = parser.parse_args(
        shlex.split(config.cmkcpiodir_opts, posix=True) + sys.argv[1:]
    )

    _set_logging_level(args.verbose, args.quiet)

    # Parse arguments
    if args.debug or args.only_build_directory:
        args.keep = True

    # Keymap
    if config.keymap is not None and config.keymap[0] is not None:
        with open(config.keymap[1], 'wb') as keymap_bin:
            mkramfs.keymap_build(
                config.keymap[0], keymap_bin,
                unicode=(locale.getdefaultlocale()[1] == 'UTF-8')
            )

    # Init
    with open(config.init_path, 'w') as init_file:
        mkinit(
            out=init_file,
            root=config.root,
            mounts=config.mounts,
            keymap=(None if config.keymap is None else config.keymap[2]),
            modules=config.modules,
            scripts=config.scripts,
        )

    # Initramfs
    if not args.only_build_archive:
        assert config.init_path is not None
        logger.info("Creating initramfs")
        initramfs = mkramfs.Initramfs(
            user=(0 if not args.debug else os.getuid()),
            group=(0 if not args.debug else os.getgid()),
            binroot=args.binroot,
            kernels=(() if args.no_kmod else args.kernel),
        )
        _build_initramfs(initramfs, config)

    if not args.only_build_archive:
        # Pre-build cleanup
        if args.clean and os.path.exists(args.build_dir):
            logger.warning("Overwriting %s", args.build_dir)
            shutil.rmtree(args.build_dir)

        # Build
        logger.info("Building initramfs to directory %s", args.build_dir)
        initramfs.build_to_directory(args.build_dir,
                                     do_nodes=(not args.debug))

    if not args.only_build_directory:
        # Create CPIO archive
        logger.info("Generating CPIO archive to %s from %s",
                    args.output, args.build_dir)
        if args.output == '-':
            mkramfs.mkcpio_from_dir(args.build_dir, sys.stdout.buffer)
        else:
            with open(args.output, 'wb') as cpiodest:
                mkramfs.mkcpio_from_dir(args.build_dir, cpiodest)

    if not args.keep:
        # Cleanup temporary files
        if os.path.exists(args.build_dir):
            logger.info("Cleaning %s", args.build_dir)
            shutil.rmtree(args.build_dir)
        if config.keymap is not None and os.path.exists(config.keymap[1]):
            logger.info("Cleaning %s", config.keymap[1])
            os.remove(config.keymap[1])
        if os.path.exists(config.init_path):
            logger.info("Cleaning %s", config.init_path)
            os.remove(config.init_path)



"""A customizable simple initramfs generator"""

__author__ = 'Louis Leseur'
__copyright__ = 'Copyright (c) 2020 Louis Leseur'
__email__ = 'louis.leseur@gmail.com'
__license__ = 'MIT'
__url__ = 'https://github.com/teapot9/fand'
__version__ = '0.2.0'

import logging

logging.basicConfig(
    format='%(levelname)s: %(message)s',
    level=logging.NOTSET,
)



"""A customizable simple initramfs generator"""

import setuptools

from cmkinitramfs import (__author__, __doc__, __email__, __license__,
                          __name__, __url__, __version__)

setuptools.setup(
    name=__name__,
    version=__version__,
    description=__doc__,
    long_description=open('README.rst', 'r').read(),
    long_description_content_type='text/x-rst',

    author=__author__,
    author_email=__email__,
    url=__url__,
    license=__license__,
    license_files=['LICENSE.txt'],

    python_requires='>=3.7, <4',
    install_requires=['pyelftools'],
    extras_require={
        'doc': ['sphinx', 'sphinx_rtd_theme'],
        'qa': ['flake8', 'mypy', 'tox'],
    },

    packages=['cmkinitramfs'],
    entry_points={
        'console_scripts': [
            'cmkinit = cmkinitramfs.entry:entry_cmkinit',
            'cmkcpiodir = cmkinitramfs.entry:entry_cmkcpiodir',
            'cmkcpiolist = cmkinitramfs.entry:entry_cmkcpiolist',
            'findlib = cmkinitramfs.entry:entry_findlib',
        ],
    },

    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: System :: Boot",
        "Topic :: System :: Boot :: Init",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
    keywords=['initramfs', 'initramfs-generator'],
    platforms=['Linux'],
)
